"""Module bulk_slab of viperleed.calc.classes.slab.

Defines the BulkSlab class, a BaseSlab describing a 3D-periodic
crystal. This module was created as part of the refactoring of
slab.py (originally created on 2019-06-13).
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-02-21'
__license__ = 'GPLv3+'

import copy
import functools
import itertools

import numpy as np
from scipy.spatial.distance import cdist as euclid_distance

from viperleed.calc.lib import leedbase
from viperleed.calc.lib.base import add_edges_and_corners, collapse
from viperleed.calc.lib.base import rotation_matrix_order

from .base_slab import BaseSlab
from .errors import AlreadyMinimalError
from .errors import MissingSublayersError
from .errors import SlabError
from .utils import _cycle


class BulkSlab(BaseSlab):
    """A class representing an infinite solid, period in 3D.

    Contains unit cell, element information and atom coordinates.
    Also has a variety of convenience functions for manipulating
    and updating the atoms.

    Attributes
    ----------
    ucell : np.array
        The unit cell as vectors a, b, c (columns)
    poscar_scaling : float
        The original scaling factor from POSCAR
    elements : tuple of str
        Element labels as read from POSCAR
    chemelem : set of str
        Chemical elements in the slab, including from `ELEMENT_MIX`
    n_per_elem : dict {str: int}
        The number of atoms per POSCAR element.
    atlist : AtomList
        List of all atoms in the slab.
    layers : tuple of Layer
        Each `layer` is a composite of sublayers, as in TensErLEED
    sublayers : tuple of SubLayer
        Each SubLayer contains atoms of equal element and Z coordinate
    sitelist : list of Sitetype
        List of distinct sites as Sitetype, storing information
        on vibration and concentration
    ucell_mod : list of tuples (str, numpy.ndarray)
        Stored modifications made to the unit cell; each is a tuple
        of (type, array), where type is 'lmul', 'rmul', 'add', or
        'c_shift'.
    topat_ori_z : float
        Stores the original position of the topmost atom in Cartesian
        coordinates
    celltype : str                                                              # TODO: would be nicer with an Enum
        Unit-cell shape as string. Values: 'oblique', 'rhombic',
        'rectangular', 'square', 'hexagonal'
    planegroup : str
        Symmetry group of the slab. May be reduced by the user
        relative to `foundplanegroup`.
    foundplanegroup : str
        Highest symmetry found. Doesn't get modified when user
        reduces symmetry manually.
    orisymplane : SymPlane
        Only stored if the `planegroup` is ambiguous as to which unit
        vector the symmetry plane at the origin is parallel to
    linklists : list of list of Atom
        List of lists of atoms which are linked by a symmetry operation
    bulk_screws : list of int
        Integer list of rotation orders present in the bulk.
    bulk_glides : list of SymPlane
        List of glide-symmetry planes present in the bulk.
    """

    def __init__(self):
        """Initialize instance."""
        super().__init__()
        del self.bulkslab
        self.bulk_screws = []
        self.bulk_glides = []

    @property
    def is_bulk(self):
        """Return whether this is a bulk slab."""
        return True

    @property
    def smallest_interlayer_gap(self):
        """Return the smallest z gap between two adjacent layers.

        Make sure to update_layer_coordinates() before.

        Returns
        -------
        min_dist : float
            The smallest of the z distances between adjacent layers.
            Distances are calculated between the bottommost atom of
            the higher layer and the topmost atom of the lower one.

        Raises
        ------
        MissingLayersError
            If no layers are available.
        """
        if self.n_layers == 1:
            return self.c_vector[2] - self.layers[0].thickness
        return super().smallest_interlayer_gap

    @classmethod
    def from_slab(cls, other):
        """Return a `cls` instance with attributes deep-copied from `other`.

        Parameters
        ----------
        other : BaseSlab
            The slab whose attributes are to be copied.

        Returns
        -------
        new_slab : SurfaceSlab
            A new slab instance with attributes copied from
            `other`. Notice that no modification will occur
            for new_slab.layers. This means that some of the
            .layers of `new_slab` may be non-bulk. The same
            holds true for atoms. The caller is responsible
            for modifying layer.is_bulk after this call.
        """
        return super().from_slab(other)

    # Disabled too-many-arguments below because 7/5 seem better than
    # packing these arguments into some data structure which would
    # make the call to this method harder to follow.
    # pylint: disable-next=too-many-arguments
    def apply_bulk_cell_reduction(self, eps, epsz=None,
                                  new_c_vec=None,
                                  new_ab_cell=None,
                                  recenter=True,
                                  z_periodic=True):
        """Reduce bulk unit cell in- and/or out-of-plane.

        Extra atoms left after reduction of the unit cell size
        are removed.

        If desired, and the unit cell is actually changed, the
        atom that is currently at the top can be kept at the top
        also later. In this case, the z position of the new unit
        cell is also centred relative to the geometric centre of
        the atom coordinates.

        Parameters
        ----------
        eps : float
            Cartesian tolerance (Angstrom) for in-plane comparison
            of atomic coordinates.
        epsz : float or None, optional
            Cartesian tolerance (Angstrom) for comparison of atomic
            coordinates in the direction orthogonal to the surface.
            If not given or None, it is taken equal to `eps`. Default
            is None.
        new_c_vec : Sequence or None, optional
            Cartesian coordinates of the new repeat vector. If not
            given or None, the current repeat vector is retained.
            Shape (3,). Default is None.
        new_ab_cell : Sequence or None, optional
            Cartesian coordinates of the new surface unit cell. The
            new unit vectors are rows, i.e., a, b = `new_ab_cell`.
            If not given or None, the current surface unit cell is
            retained. Default is None.
        recenter : bool, optional
            Whether atom coordinates along the c axis should be
            modified so that the atom that was highest when this
            method was called (or one equivalent to it upon
            translation) is also highest afterwards. Default
            is True.
        z_periodic : bool, optional
            Whether the new_c_vec assigned will make this BulkSlab
            periodic along the z direction. Default is True.

        Returns
        -------
        None.
        """
        if new_ab_cell is None and new_c_vec is None:
            return  # Nothing to do

        if epsz is None:
            epsz = eps
        update_origin = new_c_vec is not None

        # Keep track of the atom that is currently at the top:
        # after reducing the unit-cell vector along c, the fractional
        # positions are screwed. We rely on the Cartesian ones, and,
        # if requested, will 'shift' the fractional ones such that
        # the topmost atom now is still the topmost atom later
        top_atom = self.top_atom if recenter else None

        # Make sure Cartesians are up to date,
        # then reduce c direction if needed
        self.update_cartesian_from_fractional()
        if new_c_vec is not None:
            self.c_vector[:] = new_c_vec

        # Reduce in-plane dimension
        if new_ab_cell is not None:
            self.ab_cell[:] = new_ab_cell.T

        # Collapse the atom coordinates to the new cell. We collapse
        # the coordinates in a different manner if this slab is turned
        # into a periodic one: atoms close to the edges in z are
        # collapsed towards c=0. This prevents atoms closer to the
        # edges than eps/epsz from being assigned different sublayers
        # later on.
        if not z_periodic:
            self.collapse_cartesian_coordinates(update_origin=update_origin)
        else:
            # Here we do a version of collapse_cartesian_coordinates
            # that accounts for eps.
            releps = np.dot((eps, eps, epsz), np.linalg.inv(self.ucell.T))
            self.update_fractional_from_cartesian()
            self.collapse_fractional_coordinates(releps=releps)
            self.update_cartesian_from_fractional(update_origin=update_origin)

        # Get rid of duplicates resulting from the size reduction
        self.remove_duplicate_atoms(eps, epsz)

        if not recenter:
            return

        # As mentioned above, fractional positions in c
        # are now screwed. Make top atom the topmost again...
        top_atom_cfrac = top_atom.pos[2]
        for atom in self:
            atom.pos[2] = (atom.pos[2] - top_atom_cfrac + 0.9999) % 1.0
        # ...then center fractional c coordinates around cell midpoint
        atoms_c_frac = [at.pos[2] for at in self]
        midpos = (max(atoms_c_frac) + min(atoms_c_frac)) / 2
        for atom in self:
            atom.pos[2] = (atom.pos[2] - midpos + 0.5) % 1.0
        self.update_cartesian_from_fractional(update_origin=update_origin)

    def ensure_minimal_c_vector(self, rpars, z_periodic=False):
        """Reduce the c vector to its minimum value.

        This method not only finds the shortest c vector, but also
        reduces the extension of this slab to have the new c vector
        as the out-of-plane repeat. It also modifies `rpars` with
        the minimal vector found. Use `get_minimal_c_vector` if you
        only want to detect the shortest repeat, without affecting
        either the slab or `rpars`.

        Parameters
        ----------
        rpars : Rparams
            The PARAMETERS read from file. Attributes used:
            read-only: SYMMETRY_EPS.
            overwritten: BULK_REPEAT. The minimal c vector
            found is stored in the BULK_REPEAT attribute of
            `rpars`.
        z_periodic : bool, optional
            Whether the slab is to be considered periodic along
            the direction perpendicular to the surface while the
            c vector is minimized. Typically False, unless the
            current c vector is already a repeat vector for this
            slab. Default is False.

        Returns
        -------
        shortest_c : numpy.ndarray
            The new shortest c vector found.

        Raises
        ------
        AlreadyMinimalError
            If no shorter c vector is found.
        SlabError
            If this slab has too large of a vacuum gap at the bottom.
        """
        self.create_sublayers(rpars.SYMMETRY_EPS.z)

        # May raise AlreadyMinimalError
        rpars.BULK_REPEAT = self.get_minimal_c_vector(rpars.SYMMETRY_EPS,
                                                      rpars.SYMMETRY_EPS.z,
                                                      z_periodic=z_periodic)
        bottom_atom = self.bottom_atom
        delta_z = self.topat_ori_z - bottom_atom.cartpos[2]                     # TODO: flip with .cartpos[2] -- Issue #174
        if not z_periodic and delta_z > rpars.SYMMETRY_EPS.z:
            # Here we could also remove_vacuum_at_bottom ourselves, but
            # it is a bit of a pain as we would need to shift things
            # back up later in order to keep the z centering.
            raise SlabError(
                f'{type(self).__name__}.ensure_minimal_c_vector: cannot reduce'
                ' c vector for a non-z-periodic slab that has a vacuum gap at '
                f'the bottom (gap={delta_z:.3f} > {rpars.SYMMETRY_EPS.z:.3f}).'
                f' Shift all atoms down along z using '
                'slab.remove_vacuum_at_bottom(rpars)'
                )

        # Now we're sure that this slab will become
        # periodic after modifying the c vector
        self.apply_bulk_cell_reduction(rpars.SYMMETRY_EPS,
                                       epsz=rpars.SYMMETRY_EPS.z,
                                       new_c_vec=rpars.BULK_REPEAT,
                                       z_periodic=True)
        self.create_sublayers(rpars.SYMMETRY_EPS.z)
        return rpars.BULK_REPEAT

    def get_bulk_3d_str(self):
        """Return info about bulk screw axes and glide planes as a string.

        Returns
        -------
        bulk_3d_str : str
            Format is 'r(2, 4), m([1,1], [ 1,-1])' if there is
            any screw axes or glide planes, otherwise 'None'.
        """
        return leedbase.bulk_3d_string(self.bulk_screws,
                                       (p.par for p in self.bulk_glides))

    def get_bulk_repeat(self, rpars, only_z_distance=False):
        """Return the bulk repeat vector (with positive z).

        Parameters
        ----------
        rpars : Rparams
            The PARAMETERS to be interpreted. This is unused for
            a bulk slab.
        only_z_distance : bool, optional
            Whether a distance in the direction perpendicular to the
            surface (i.e., not necessarily along the c axis) should
            be returned rather than a full vector. This is ignored
            if `rpars.BULK_REPEAT` is a vector. Default is False.

        Returns
        -------
        bulk_repeat_vector : numpy.ndarray or float
            Bulk repeat vector pointing from the bulk to the surface,
            or its component along z.
        """
        bulkc = self.c_vector.copy()
        if bulkc[2] < 0:
            bulkc *= -1

        if isinstance(rpars.BULK_REPEAT, np.ndarray) or not only_z_distance:
            return bulkc
        return bulkc[2]

    def get_candidate_layer_periods(self, epsz=1e-4):
        """Find at which offsets the sublayers repeat in a bulk slab.

        Parameters
        ----------
        epsz : float, optional
            Tolerance (Cartesian) on z distances. Default is 1e-4.

        Returns
        -------
        candidate_periods : list
            Each element is an integer corresponding to the
            number of sublayers that potentially constitute
            a period for the slab. They are 'potential' in
            that only consistency of chemical elements and
            number of atoms are considered.

        Raises
        ------
        MissingSublayersError
            If this slab has no sublayers defined.
        """
        if not self.sublayers:
            raise MissingSublayersError
        if self.n_sublayers < 2:
            return []

        # Start from one layer (e.g., the first one) and accumulate
        # potential periods as the offsets of all other layers with
        # same element and same number of atoms. Stop at half the
        # unit cell height, as periodicity cannot go beyond h/2.
        # The same is true for the number of sublayers.
        n_layers = self.n_sublayers
        candidate_periods = []
        ucell_h = self.c_vector[2]
        ref_lay = self.sublayers[0]
        for i, lay in enumerate(self.sublayers[1:], start=1):
            if (abs(lay.cartbotz - ref_lay.cartbotz) > ucell_h / 2 + epsz
                    or i > n_layers / 2):
                break
            if (lay.element == ref_lay.element
                    and lay.n_atoms == ref_lay.n_atoms):
                candidate_periods.append(i)

        # Now make sure that also the layers
        # in between have the same periods
        return [p for p in candidate_periods
                if self._is_candidate_period_acceptable(p, epsz)]

    def _is_candidate_period_acceptable(self, period, epsz):
        """Return whether a sublayer period is an acceptable candidate.

        Parameters
        ----------
        period : int
            Index distance between sublayers to be checked. This method
            assumes that self.sublayers[0] and self.sublayers[period]
            are consistent (same element and number of atoms).
        epsz : float
            Tolerance (Cartesian) on z distances.

        Returns
        -------
        acceptable : bool
            Whether `period` is a suitable sublayer period. This
            means that all the pairs of sublayers of the form
            (self.sublayers[0<i<period], self.sublayers[i+period])
            match one another (same element and number of atoms),
            and have the same z distance as the one between
            self.sublayers[0] and self.sublayers[period].
        """
        n_layers = self.n_sublayers
        ucell_h = self.c_vector[2]

        # Keep track of the distance between layers that
        # we know match: the first one and the one at period.
        # All pairs of layers in between must also be at the
        # same distance.
        z_dist = self.sublayers[period].cartbotz - self.sublayers[0].cartbotz
        layer_pairs = zip(self.sublayers, _cycle(self.sublayers, period))
        for i, (lay, lay_plus_period) in enumerate(layer_pairs, start=1):
            if i > n_layers / 2:
                return True
            z_dist_this = lay_plus_period.cartbotz - lay.cartbotz
            z_dist_this %= ucell_h
            if (lay_plus_period.element != lay.element
                    or lay_plus_period.n_atoms != lay.n_atoms
                    or abs(z_dist - z_dist_this) > epsz):
                return False
        return True

    def get_minimal_c_vector(self, eps, epsz=None, z_periodic=True):
        """Return the smallest bulk c vector, if any.

        Notice that this method does not modify this slab. If
        you want this slab to have the shortest c vector, use
        `ensure_minimal_c_vector` instead.

        Parameters
        ----------
        eps : float
            Cartesian tolerance for in-plane distances
        epsz : float or None, optional
            Cartesian tolerance for distances in the z direction,
            i.e., perpendicular to the surface. If not given or
            None, it is taken equal to `eps`.
        z_periodic : bool, optional
            Whether the current slab should be considered periodic
            in the direction of the c vector. This is commonly True
            for all bulk slabs, unless the current c vector is not
            a repeat vector (e.g., while c is being identified using
            detect_bulk). Default is True.

        Returns
        -------
        shortest_c : numpy.ndarray
            The minimal repeat c vector found, in Cartesian
            coordinates with (x, y) in the surface plane, and
            z directed from the solid towards the surface, i.e.,
            opposite to the usual LEED convention. The in-plane
            components are always minimized.

        Raises
        ------
        AlreadyMinimalError
            If a shorter repeat vector could not be found.
        """
        if epsz is None:
            epsz = eps
        periods = self.get_candidate_layer_periods(epsz)
        if not periods:
            raise AlreadyMinimalError(
                f'{type(self).__name__}.get_minimal_c_vector: no '
                'candidate sublayer periods for reducing c vector'
                )

        # Work with a deepcopy, as we will play around with sublayers
        self_copy = copy.deepcopy(self)
        self_copy.update_cartesian_from_fractional()
        self_copy.create_sublayers(epsz)
        n_layers = self_copy.n_sublayers

        # Pick one layer (e.g., the first one) and test translations
        # for all vectors connecting an atom of this layer (e.g., the
        # first one) to all atoms of all other layers.
        ori = self_copy.sublayers[0].cartpos

        # Since the periods are sorted from smaller to larger, we need
        # only keep track of one repeat vector, stopping at the earliest
        _is_symmetric = functools.partial(self_copy.is_translation_symmetric,
                                          eps=eps, z_periodic=z_periodic)
        repeat_c = None
        for period in periods:
            other_layer = self_copy.sublayers[period % n_layers]
            # Vectors from surface to bulk (the way we want them
            # later), but flip them for testing matching, which
            # means translating the slab up towards the surface.
            # This is important when, e.g., we're looking for the
            # c vector in SurfaceSlab.detect_bulk: the upper part
            # of the slab is not bulk.
            test_vecs = (ori - atom.cartpos for atom in other_layer)
            repeat_c = next((v for v in test_vecs if _is_symmetric(-v)), None)
            if repeat_c is not None:
                break

        if repeat_c is None:
            raise AlreadyMinimalError(
                f'{type(self).__name__}.get_minimal_c_vector: none '
                'of the candidate sublayer periods is an actual period'
                )

        # Flip z, because we store it opposed in Atom.cartpos[2], but
        # want repeat_c with the same coordinate system as unit cell
        repeat_c[2] *= -1                                                       # TODO: .cartpos[2] -- Issue #174

        # Minkowski-reduce the repeat vector to have
        # it shortest and as close to z as possible
        leedbase.reduce_c_vector(repeat_c, self.ab_cell.T)
        return repeat_c

    def is_bulk_glide_symmetric(self, symplane, sublayer_period, eps):
        """Return if the slab has a 3D glide plane.

        Parameters
        ----------
        symplane : SymPlane
            The mirror part of the glide plane to test.
        sublayer_period : int
            Number of sublayers, corresponding to the translation
            part of the glide operation to be tested: translations
            are tested for all vectors connecting pairs of sublayers
            whose indices differ by `sublayer_period`. Normally one
            of the values returned by `get_candidate_layer_periods`.
        eps : float
            Tolerance (Cartesian) on atomic positions.

        Returns
        -------
        has_glide : bool
            Whether the slab has a bulk glide plane with the
            given `symplane` and `sublayer_period` translation.
        """
        matrix = symplane.point_operation(n_dim=3)
        return self._is_bulk_transform_symmetric(matrix, sublayer_period, eps)

    def is_bulk_screw_symmetric(self, order, sublayer_period, eps):
        """Return if the slab has a 3D screw axis.

        Parameters
        ----------
        order : int
            The rotation order of the screw axis to test.
        sublayer_period : int
            Number of sublayers, corresponding to the translation
            part of the screw operation to be tested: translations
            are tested for all vectors connecting pairs of sublayers
            whose indices differ by `sublayer_period`. Normally one
            of the values returned by `get_candidate_layer_periods`.
        eps : float
            Tolerance (Cartesian) on atomic positions.

        Returns
        -------
        has_screw : bool
            Whether the slab has a bulk screw with the given
            `order` and `sublayer_period` translation.
        """
        matrix = rotation_matrix_order(order, dim=3)
        return self._is_bulk_transform_symmetric(matrix, sublayer_period, eps)

    # Disabled too-many-locals below: While there are indeed quite
    # a few locals (20/15), refactoring this would require either
    # helper methods that take a lot of arguments or some special
    # class that holds them. That's because all the stuff is computed
    # the very least number of times to speed up execution. Let's
    # prefer execution time and use a lot of comments to help.
    # pylint: disable-next=too-many-locals
    def _is_bulk_transform_symmetric(self, matrix, sublayer_period, eps):
        """Return if this slab is equivalent under a 3D screw/glide.

        Parameters
        ----------
        matrix : numpy.ndarray
            The rotational or mirror part of the symmetry
            transformation to be tested. Shape (3, 3). The
            transformation `matrix` is applied from the left to
            the Cartesian coordinates (taken as column vectors).
            It should represent a point operation that does not
            change the z coordinates, i.e., it is block-diagonal
            with non-trivial elements only in [:2, :2].
        sublayer_period : int
            Number of sublayers to be considered for constructing
            test translation vectors. Should be one of the candidate
            periods returned by `self.get_candidate_layer_periods()`.
            Translational symmetry is tested for all vectors connecting
            atoms between sublayer pairs whose indices differ by
            `sublayer_period`.
        eps : float
            Tolerance (Cartesian) for position equivalence.

        Returns
        -------
        is_symmetric : bool
            True if screw/glide symmetric, else False.
        """
        # Let's use the better version of the unit cell, with
        # unit vector as rows
        ucell = self.ucell.T
        ucell_inv = np.linalg.inv(ucell)
        releps = eps / np.linalg.norm(ucell, axis=1)

        # Get translation vectors to check. Notice that it is enough
        # to pick any single atom from any layer as 'reference', and
        # test translations to all atoms in another layer. To minimize
        # the number of test translations, use pairs of low-occupancy
        # layers. The next lines assume that sublayer_period comes from
        # a call to get_candidate_layer_periods.
        lowocclayer = self.fewest_atoms_sublayer
        ori = matrix.dot(lowocclayer.cartpos)
        ref_layers = itertools.islice(self.sublayers,  # From next one
                                      lowocclayer.num + 1, None)
        # compare_layers: sublayers starting at num+period index,
        # wrapped around. Notice that num+period is the buddy of
        # lowocclayer. We take it out of the iterator while making
        # the translations. After then, compare_layers is aligned
        # with ref_layers.
        compare_layers = _cycle(self.sublayers,
                                lowocclayer.num + sublayer_period)
        translations = [at.cartpos - ori for at in next(compare_layers)]

        for ref_layer, compare_layer in zip(ref_layers, compare_layers):
            # Prepare matrix-transformed versions of the Cartesian
            # coordinates of this sublayer, after collapsing to the
            # base unit cell. Translations will be applied later.
            matrix_transformed, _ = collapse(
                np.array([at.cartpos for at in ref_layer]),
                ucell, ucell_inv
                )
            matrix_transformed = matrix_transformed.dot(matrix.T)

            # Get coordinates of the sublayer to compare to, also
            # collapsed to base cell, and including extra atoms
            # for those close to edges and corners
            to_compare, frac_coords = collapse(
                np.array([at.cartpos for at in compare_layer]),
                ucell, ucell_inv
                )
            to_compare, _ = add_edges_and_corners(to_compare,
                                                  frac_coords,
                                                  releps, ucell)
            j = 0
            while j < len(translations):
                transformed_3d = matrix_transformed + translations[j]
                transformed_3d, _ = collapse(transformed_3d, ucell, ucell_inv)
                distances = euclid_distance(transformed_3d, to_compare)
                if any(distances.min(axis=1) > eps):
                    translations.pop(j)
                else:
                    j += 1
            if not translations:
                return False
        return True

    def with_double_thickness(self, new_atoms_start_idx=None):
        """Return a copy of this bulk slab which is twice as thick."""
        double_slab = copy.deepcopy(self)
        c_vec = double_slab.c_vector

        # For atoms that are added, because we use z flipped                    # TODO: .cartpos[2] -- Issue #174
        c_vec_atoms = c_vec.copy()
        c_vec_atoms[2] *= -1
        double_slab.update_cartesian_from_fractional(update_origin=True)

        if new_atoms_start_idx is None:
            # BulkSlab objects tend to have a non-continuous
            # distribution of atom numbers that usually come
            # from the parent SurfaceSlab for which this slab
            # is the bulk. We cannot use the normal n_atoms + 1,
            # as we may end up in a conflict of atom numbers.
            new_atoms_start_idx = max(at.num for at in double_slab) + 1

        # Now decide which way to go, depending on
        # whether there are layers already defined
        if double_slab.layers:
            # pylint: disable=protected-access
            double_slab._add_one_bulk_cell(double_slab.layers, c_vec,
                                           c_vec_atoms, 0, new_atoms_start_idx)
        else:
            for atom in double_slab.atlist.copy():
                atom.duplicate(num=new_atoms_start_idx)
                atom.cartpos += c_vec_atoms
                new_atoms_start_idx += 1
            c_vec[:] *= 2
        double_slab.collapse_cartesian_coordinates(update_origin=True)
        double_slab.sublayers = ()  # They are outdated
        return double_slab



"""ViPErLEED utility: Project c vector to z axis.

This utility takes a POSCAR slab and projects the c vector to the z axis.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from viperleed.utilities.poscar.base import _PoscarStreamCLI


class ProjectCToZCLI(_PoscarStreamCLI, cli_name='project_c_to_z'):
    """Remove atoms between a pair of c fractions."""

    long_name = 'project c to z axis'

    def process_slab(self, slab, args):
        """Project c vector of slab to z axis."""
        slab.project_c_to_z()
        return slab


if __name__ == '__main__':
    ProjectCToZCLI.run_as_script()



"""Module cli of viperleed.

Defines the main command-line interface for running viperleed
and its subpackages.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-01'
__license__ = 'GPLv3+'

from viperleed.cli_base import ViPErLEEDCLI


class ViPErLEEDMain(ViPErLEEDCLI, cli_name='viperleed'):
    """The main CLI interface of viperleed."""

    def __init__(self, *args, **kwargs):
        """Initialize instance by registering sub-utilities."""
        super().__init__(*args, **kwargs)
        children = (
            # Notice that, because of a current fuckery with the way
            # the imports are in guilib, viperleed.gui should be the
            # first one, so that it is not imported later. Otherwise,
            # "python viperleed gui" thinks it is in command-line
            # mode, while "python viperleed.gui" works fine.
            'viperleed.gui',                                                    # TODO: gui arguments
            'viperleed.calc.bookkeeper',
            'viperleed.calc.cli',
            'viperleed.utilities.cli',
            'viperleed.utilities.poscar.cli',
            )
        self.register_valid_children(children)
        self.add_child_aliases('utilities', 'utils', 'util')



"""Module write of viperleed.calc.files.parameters.

Initial version by @fkraushofer in 2020, major rewrite by @amimre
and @michele-riva in June 2023. This module used to be part of
parameters.py. Refactored in October 2023.

Functions for writing and editing a PARAMETERS file.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-18'
__license__ = 'GPLv3+'

from collections.abc import Sequence
from contextlib import AbstractContextManager
import logging
from pathlib import Path
import shutil

import numpy as np

from viperleed.calc.lib.base import parent_name
from viperleed.calc.lib.base import strip_comments
from viperleed.calc.lib.woods_notation import writeWoodsNotation

from .reader import RawLineParametersReader


_LOGGER = logging.getLogger(parent_name(__name__))


def comment_out(rpars, modpar, comment='', path='', suppress_ori=False):
    """Comment out modpar in the PARAMETERS file."""
    editor = ParametersFileEditor(
        rpars, path=path,
        save_existing_parameters_file=not suppress_ori
        )
    with editor:
        editor.comment_out_parameter(modpar, comment=comment)


def modify(rpars, modpar, new=None, comment='', path='', suppress_ori=False):
    """Change the value for `modpar` in the PARAMETERS file.

    The lines that contains `modpar` are commented out, and
    replaced by the string specified by `new`, or the value
    currently stored in `rpars`.

    Parameters
    ----------
    rpars : Rparams
        The run parameters object.
    modpar : str
        The parameter that should be modified.
    new : object or None, optional
        The new value for the parameter. If None, the current
        value of `modpar` in `rpars` is used instead. Default
        is None.
    comment : str, optional
        A comment to be added in the new line in PARAMETERS.
    path : str or Path, optional
        Where to find the PARAMETERS file that should be modified.
        Default is an empty string, i.e., the current directory.
    suppress_ori : bool, optional
        If True, no 'PARAMETERS_original' file will be created.

    Returns
    -------
    new_value : str
        String value of the parameter as inserted in the
        PARAMETERS file.
    """
    editor = ParametersFileEditor(
        rpars, path=path,
        save_existing_parameters_file=not suppress_ori
        )
    with editor:
        new_param = editor.modify_param(modpar, new_value=new, comment=comment)
    return new_param.fmt_value


# This is almost a dataclass (but not quite), all the
# information is static and generated only once via
# private methods. We don't really need public ones.
# pylint: disable-next=too-few-public-methods
class ModifiedParameterValue:
    """A container for a single modified PARAMETER.

    Attributes
    ----------
    already_written : bool
        Whether this parameter was already written to file.
    comment : str
        An optional comment that should accompany this
        modified parameter when it will be written to file
    fmt_value : str
        A formatted version of the new value. This normally
        includes only the part on the right side of the equals
        sign. `fmt_value` is the empty string if this parameter
        should be `only_comment_out`.
    line : str
        A formatted version of the full line to be written
        to file when modifying the parameter. This contains
        also the comment, if given, and is `\n`-terminated.
        `line` is the empty string if the parameter should
        be `only_comment_out`.
    only_comment_out : bool
        Whether this parameter should only be commented out.
    param : str
        The PARAMETERS parameter to be modified.
    """

    _simple_params = {   # Those with a specific format string
        'BULK_LIKE_BELOW': '.4f',
        'LAYER_CUTS': '.4f',
        'V0_IMAG': '.4f',
        }
    _string_params = {   # Those for which str is enough
        'LMAX',
        'N_BULK_LAYERS',
        'SYMMETRY_FIX',  # For now
        }
    _vector_params = {   # (fmt_for_element, brackets)
        'THEO_ENERGIES': ('{:.4g}', ''),
        }
    _wood_or_matrix = {'SYMMETRY_CELL_TRANSFORM', 'SUPERLATTICE'}

    def __init__(self, param, rpars_or_value,
                 comment='', only_comment_out=False):
        """Initialize instance.

        Parameters
        ----------
        param : str
            The PARAMETERS parameter whose value is modified.
        rpars_or_value : object
            If an Rparams object, the new value is taken from
            the current value of the corresponding attribute,
            otherwise, it is treated as the raw version of the
            new value.
        comment : str, optional
            Extra comments to add to the line that should be
            written.
        only_comment_out : bool, optional
            Whether the parameter should only be removed by
            commenting, and not actually modified.

        Returns
        -------
        None.
        """
        self.only_comment_out = only_comment_out
        self.already_written = False
        self.param = param
        self.comment = comment
        self._raw_value = None
        self.fmt_value = ''
        self.line = ''
        if not self.only_comment_out:
            self._update(rpars_or_value)

    def _update(self, rpars_or_value):
        """Gather the value, its formatted version and a line."""
        try:
            rpars_or_value.BULK_REPEAT
        except AttributeError:  # Assume it's a value
            self._raw_value = rpars_or_value
        else:                   # Assume it's an Rparams
            self._raw_value = self._get_raw_value(rpars_or_value)
        self.fmt_value = self._format_value()
        self.line = self._format_line()

    def _format_value(self):
        """Return a formatted version of raw_value for param."""
        param = self.param
        if param in self._string_params:
            return str(self._raw_value)

        fmt_string = self._simple_params.get(param, None)
        if fmt_string:
            return f'{self._raw_value:{fmt_string}}'

        if param in self._wood_or_matrix:
            return self._format_wood_or_matrix_value()

        if param in self._vector_params:
            args = self._vector_params[param]
            return self._format_vector(self._raw_value, *args)

        try:
            formatter = getattr(self, f'_format_{param.lower()}_value')
        except AttributeError:
            raise ValueError(f'No formatting available for {param}') from None
        return formatter()

    def _format_line(self):
        """Return a formatted version of the new line for a PARAMETER."""
        value = self.fmt_value
        line = value if '=' in value else f'{self.param} = {value}'
        if self.comment:
            line = f'{line.rstrip():<35} ! {self.comment}'
        return line + '\n'

    def _format_beam_incidence_value(self):
        """Return a formatted version of BEAM_INCIDENCE."""
        theta, phi = self._raw_value
        return f'THETA {theta:.4f}, PHI {phi:.4f}'

    def _format_bulk_repeat_value(self):
        """Return a formatted version of BULK_REPEAT."""
        bulk_repeat = self._raw_value
        if isinstance(bulk_repeat, (np.ndarray, Sequence)):
            return self._format_vector(bulk_repeat, '{:.5f}', brackets='[]')
        return f'{bulk_repeat:.5f}'  # Z distance

    def _format_wood_or_matrix_value(self):
        """Return a formatted version of an integer matrix value."""
        formatted = writeWoodsNotation(self._raw_value)
        if formatted:
            formatted = f'= {formatted}'
        else:
            matrix = self._raw_value.round().astype(int)
            # Disable pylint, as the *ravel way looks cleaner
            # than an f-string with explicit matrix elements
            # pylint: disable-next=consider-using-f-string
            formatted = 'M = {} {}, {} {}'.format(*matrix.ravel())
        return f'{self.param} {formatted}'

    @staticmethod
    def _format_vector(vector, elem_fmt, brackets=''):
        """Return a formatted version of a 1D vector."""
        elements = ' '.join(elem_fmt.format(v) for v in vector)
        if not brackets:
            return elements
        left, right = brackets
        return left + elements + right

    def _get_raw_value(self, rpars):
        """Return the current value of a parameter from rpars."""
        try:
            return getattr(rpars, self.param)
        except AttributeError:
            pass
        getter = getattr(self, f'_get_raw_value_{self.param.lower()}', None)
        if getter is None:
            raise ValueError(f'No raw-value getter for {self.param}')
        return getter(rpars)

    @staticmethod
    def _get_raw_value_beam_incidence(rpars):
        """Return theta and phi."""
        return rpars.THETA, rpars.PHI


class ParametersFileEditor(AbstractContextManager):
    """A context manager for editing a PARAMETERS file."""

    _filename = 'PARAMETERS'
    _header = """

! ######################################################
! #  THE FOLLOWING LINES WERE GENERATED AUTOMATICALLY  #
! ######################################################

"""

    def __init__(self, rpars, path='', save_existing_parameters_file=True):
        """Initialize instance."""
        self._path = Path(path).resolve()
        self._rpars = rpars
        self._should_save_existing_file = save_existing_parameters_file

        self._to_modify = {}  # Parameters to modify/comment out
        self._has_header = False
        self._write_param_file = None  # The file object to write to

    @property
    def _file(self):
        """Return the path to the PARAMETERS file to modify."""
        return self._path / self._filename

    def __enter__(self):
        """Prepare to modify a PARAMETERS file."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Modify PARAMETERS file before exiting."""
        self.write_modified_parameters()
        return super().__exit__(exc_type, exc_value, traceback)

    def comment_out_parameter(self, param, comment=''):
        """Mark param as to be completely commented out."""
        commented = ModifiedParameterValue(param, self._rpars,
                                           comment=comment,
                                           only_comment_out=True)
        self._to_modify[param] = commented
        return commented

    def modify_param(self, param, new_value=None, comment=''):
        """Mark param as to be modified from the current value in rpars."""
        if new_value is None:
            new_value = self._rpars
        mod_param = ModifiedParameterValue(param, new_value, comment=comment)
        self._to_modify[param] = mod_param
        return mod_param

    def save_existing_parameters_file(self):
        """Save a copy of an existing PARAMETERS file as 'ori_<timestamp>'."""
        if not self._should_save_existing_file:
            return

        oriname = f'{self._filename}_ori_{self._rpars.timestamp}'
        if not self._file.is_file() or oriname in self._rpars.manifest:
            self._should_save_existing_file = False
            return

        try:
            shutil.copy2(self._file, self._path / oriname)
        except OSError:
            _LOGGER.error(
                f'parameters.modify: Could not copy {self._filename} file to '
                f'{self._filename}_ori. Proceeding. Original file will be lost'
                )
        self._rpars.manifest.append(oriname)
        self._should_save_existing_file = False

    def write_modified_parameters(self):
        """Write a new PARAMETERS, modifying all the requested lines."""
        self.save_existing_parameters_file()
        cwd = Path().resolve()
        if self._filename not in self._rpars.manifest and self._path == cwd:
            self._rpars.manifest.append(self._filename)

        if not self._to_modify:
            return

        # Collect the contents of the PARAMETERS
        # file before re-opening it for writing
        lines = self._collect_lines()
        self._has_header = False
        _head_mark = '! #  THE FOLLOWING LINES WERE GENERATED AUTOMATICALLY  #'
        with self._file.open('w', encoding='utf-8') as self._write_param_file:
            for param, raw_line in lines:
                if _head_mark in raw_line:
                    self._has_header = True
                self._write_one_line(param, raw_line)
            self._write_missing_parameters()
        self._to_modify.clear()

    def _collect_lines(self):
        """Return lines read from the PARAMETERS file."""
        if not self._file.is_file():
            return ()

        reader = RawLineParametersReader(self._file, noisy=False)
        try:  # pylint: disable=too-many-try-statements
            with reader:
                return tuple(reader)
        except Exception:
            _LOGGER.error(f'Error reading {self._filename} file.')
            raise

    def _get_comment_line_for(self, modified, raw_line):
        """Return a commented version of `raw_line`."""
        if modified.only_comment_out:
            comment = modified.comment or 'line commented out automatically'
            return f'! {raw_line.rstrip():<33} ! {comment}\n'
        return f'! {raw_line.rstrip():<33} ! line automatically changed to:\n'

    def _is_unchanged(self, modified, raw_line):
        """Return whether `modified` is already present in `raw_line`."""
        return strip_comments(modified.line) == strip_comments(raw_line)

    def _write_one_line(self, param, raw_line):
        """Write one `raw_line` containing param."""
        if not param:  # Empty, comment, or invalid line
            self._write_param_file.write(raw_line)
            return

        modified = self._to_modify.get(param, None)
        if modified is None or self._is_unchanged(modified, raw_line):
            # No modification needed
            self._write_param_file.write(raw_line)
            return

        commented_line = self._get_comment_line_for(modified, raw_line)
        self._write_param_file.write(commented_line)
        self._write_new_parameter_line(modified)

    def _write_new_parameter_line(self, modified):
        """Write one line for `modified`, if necessary."""
        if modified.only_comment_out or modified.already_written:
            return
        self._write_param_file.write(modified.line)
        modified.already_written = True

    def _write_missing_parameters(self):
        """Write all the parameters that were not found."""
        to_be_written = [p for p in self._to_modify.values()
                         if not p.already_written]
        if not to_be_written:
            return
        if not self._has_header:
            self._write_param_file.write(self._header)
        self._write_param_file.writelines(p.line for p in to_be_written)



"""Functions for reading and writing PatternInfo.tlm."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import logging


logger = logging.getLogger(__name__)


def writePatternInfo(sl, rp, filename="PatternInfo.tlm"):
    """Writes a PatternInfo file that can be used by the ViPErLEED GUI utility
    to display the expected LEED pattern and show beam labelling."""
    output = f"eMax = {rp.THEO_ENERGIES.max:.2f}\n"
    mstring = "[[{}, {}], [{}, {}]]".format(*sl.ab_cell.T.ravel())
    output += "surfBasis = "+mstring+"\n"
    mstring = ("[[{:.0f}, {:.0f}], [{:.0f}, {:.0f}]]"
               .format(rp.SUPERLATTICE[0, 0], rp.SUPERLATTICE[0, 1],
                       rp.SUPERLATTICE[1, 0], rp.SUPERLATTICE[1, 1]))
    output += "superlattice = "+mstring+"\n"
    if sl.planegroup in ["pm", "pg", "cm", "rcm", "pmg"]:
        pgstring = sl.planegroup+str(sl.orisymplane.par)
    else:
        pgstring = sl.planegroup
    output += "surfGroup = "+pgstring+"\n"
    if sl.bulkslab is None:
        logger.error("PatternInfo.tlm: bulk slab has not been initialized.")
        raise RuntimeError("writePatternInfo called without bulk slab.")
    output += "bulkGroup = "+sl.bulkslab.foundplanegroup+"\n"
    output += "bulk3Dsym = "+sl.bulkslab.get_bulk_3d_str()
    # write output
    try:
        with open(filename, 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error("Failed to write "+filename)
        raise
    logger.debug("Wrote to "+filename+" successfully")
    return



"""ViPErLEED utility: Preparing POSCAR for VASP relaxation

This utility takes a slab in POSCAR format as used by ViPErLEED and prepares it
for relaxation in VASP. This includes adding a vacuum gap on top of the slab,
writing the "Selective dynamics" flags, and giving logical flags for each atom.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-02'
__license__ = 'GPLv3+'

from viperleed.calc.files import poscar
from viperleed.cli_base import float_in_zero_one
from viperleed.utilities.poscar.base import _PoscarStreamCLI

# TODO: add an option to add a mirror image to the slab, so that the slab is
#       symmetric with respect to the center of the slab. This could be useful
#       when dealing with a polar surface.

MIN_VACUUM_GAP = 10  # angstrom. Emit warnings if too small.


class PrepareForVASPRelaxCLI(_PoscarStreamCLI, cli_name='vasp_relax'):
    """Add VASP relaxation tags to the atoms of a POSCAR."""

    long_name = 'preparing slab for VASP relaxation'

    def add_parser_arguments(self, parser):
        """Add arguments specifying which atoms/directions to mark."""
        super().add_parser_arguments(parser)
        parser.add_argument(
            'above_c',
            help='specify above which c fraction to relax the slab.',
            type=float_in_zero_one,
            )
        parser.add_argument(
            '--all_directions',
            help='relax all directions, not just the c direction.',
            action='store_true'
            )

    def process_slab(self, slab, args):
        """Return an unchanged slab, as we only edit comments."""
        logger = self.get_logger()
        logger.debug(f'Relaxing above c fraction {args.above_c}.')
        vacuum = slab.vacuum_gap
        if vacuum < MIN_VACUUM_GAP:
            logger.warning(
                f'Vacuum gap between top and bottom of slab is {vacuum:.2f}A, '
                f'i.e., less than {MIN_VACUUM_GAP}A. This may result in '
                'interaction between the periodic replicas of slabs.'
                )
        return slab

    def write_output(self, processed_slab, args):
        """Write VASP POSCAR to args.outfile (typically the terminal)."""
        write_vasp_poscar(processed_slab, args)


def write_vasp_poscar(slab, args):
    """Pipe a Slab to stdout given some command-line arguments."""
    # What follows is very similar to poscar.write. The reason not
    # to do this there is to prevent adding a dedicated argument
    # that would only be used in this specific use case. It would
    # also complicate uselessly the code: it would need to decide
    # to use a VASPPOSCARWriter rather than a POSCARFileWriter
    slab.sort_by_element()
    relax_info = {'above_c': args.above_c,
                  'c_only': not args.all_directions}
    writer = poscar.VASPPOSCARWriter(args.outfile, relax_info=relax_info)
    writer.write(slab)


if __name__ == '__main__':
    PrepareForVASPRelaxCLI.run_as_script()



"""
===============================================
      ViPErLEED Graphical User Interface
===============================================
 *** module guilib.leedsim.NewFileDialog ***

Created: 2020-01-11
Author: Michele Riva

"""

import re

import numpy as np
import PyQt5.QtCore as qtc
import PyQt5.QtGui as qtg
import PyQt5.QtWidgets as qtw

from viperleed import guilib as gl

angstrom = ' \u212b'
degrees = '\u00b0'

class NewFileDialog(qtw.QDialog):
    dialogEdited = qtc.pyqtSignal(dict) # a dictionary of the LEED parameters
    
    def __init__(self, parent=None, oldParams=dict()):
        # oldParams is used to restore the old pattern if the users presses a 
        # 'cancel' button, and to initialize the values of the current 
        # parameters
        #super(NewFileDialog, self).__init__(parent)
        super().__init__(parent)
        self.setWindowModality(qtc.Qt.WindowModal)
        flags = self.windowFlags()
        flags &= ~qtc.Qt.WindowCloseButtonHint #disable close button
        self.setWindowFlags(flags)
        self.setWindowTitle('Create new LEED input')
        
        self.compose()
        self.woodL = gl.Woods()
        self.bulk_3d_sym_dialog = gl.Bulk3DSymDialog(self)
        self.open(oldParams)
    
    def open(self, params=dict()):
        #---- update some quantities for this session ----#
        self.oldParams = params  # LEED parameters when opened
        
        self.initLattices()
        self.initControlValues()

        super().open()
    
    def initLattices(self):
        if self.oldParams:  # dictionary not empty
            self.supMatrix = self.oldParams['SUPERLATTICE']
            surfBasis = self.oldParams['surfBasis']
            surfGroup = self.oldParams['surfGroup']
            bulkGroup = self.oldParams['bulkGroup']
        else:
            self.supMatrix = np.array([[1, 0], [0, 1]])
            a = [2, 0]
            b = [3*np.cos(np.radians(100)), 3*np.sin(np.radians(100))]
            surfBasis = np.array([a, b])
            surfGroup = 'p1'
            bulkGroup = surfGroup
        
        bulkBasis = np.dot(np.linalg.inv(self.supMatrix), surfBasis)
        
        self.surfLatt = gl.Lattice(surfBasis, group=surfGroup)
        self.bulkLatt = gl.Lattice(bulkBasis, group=bulkGroup)
        
        if self.oldParams:
            self.bulkLatt.group.screws_glides = (self.oldParams['bulk3Dsym'],
                                                 self.bulkLatt.cell_shape)

    def compose(self):
        labFont = gl.AllGUIFonts().labelFont
        bigLabFont = qtg.QFont(gl.AllGUIFonts().largeTextFont)
        bigLabFont.setPointSize(11)
        smallText = gl.AllGUIFonts().smallTextFont
        
        #-------- Initialize the widgets --------#
        # here only the immutable values are set #
        
        # Input type selection
        lattInputLabel = qtw.QLabel('Lattice unit cells')
        lattInputLabel.setFont(bigLabFont)
        allWidgs = [lattInputLabel]
        
        # radio buttons for type selection
        self.bulkInput = qtw.QRadioButton('Bulk')
        self.bulkInput.setFont(smallText)
        self.surfInput = qtw.QRadioButton('Surface')
        self.surfInput.setFont(smallText)
        allWidgs.extend([self.bulkInput, self.surfInput])
        
        # lattice shape drop down menus
        latShapeLabel = qtw.QLabel('Shape')
        latShapeLabel.setFont(labFont)
        self.bulkShape = qtw.QComboBox()
        self.surfShape = qtw.QComboBox()
        for combo in [self.bulkShape, self.surfShape]:
            combo.setFont(smallText)
            combo.addItems(['Oblique', 'Rectangular', 'Square',
                            'Rhombic', 'Hexagonal'])
            combo.setSizePolicy(qtw.QSizePolicy.Fixed,
                                qtw.QSizePolicy.Preferred)
            combo.adjustSize()
        allWidgs.extend([self.bulkShape, self.surfShape])
        
        # lattice parameters
        aLab = qtw.QLabel('a = ')
        bLab = qtw.QLabel('b = ')
        alphaLab = qtw.QLabel('\u03b1 = ')
        aLab.setFont(labFont)
        bLab.setFont(labFont)
        alphaLab.setFont(labFont)
        lattLabels = [aLab, bLab, alphaLab, latShapeLabel]
        [lab.adjustSize() for lab in lattLabels]
        w = max([lab.width() for lab in lattLabels])
        [lab.setMaximumWidth(w) for lab in lattLabels]
        
        self.aB = qtw.QLineEdit('')
        self.bB = qtw.QLineEdit('')
        self.alphaB = qtw.QLineEdit('')
        
        self.aS = qtw.QLineEdit('')
        self.bS = qtw.QLineEdit('')
        self.alphaS = qtw.QLineEdit('')
        self.bulkParams = [self.aB, self.bB, self.alphaB]
        self.surfParams = [self.aS, self.bS, self.alphaS]
        lattParBoxes = [*self.bulkParams, *self.surfParams]
        for tbox in lattParBoxes:
            tbox.setFont(labFont)
            tbox.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Fixed)
            tbox.setMaximumWidth(self.bulkShape.width())
        
        allWidgs.extend([*lattLabels, *lattParBoxes])
        
        # groups
        groupLab = qtw.QLabel('Group')
        groupLab.setFont(labFont)
        self.bulkGroup = qtw.QComboBox()
        self.surfGroup = qtw.QComboBox()
        for combo in [self.bulkGroup, self.surfGroup]:
            combo.setFont(smallText)
            combo.setSizePolicy(qtw.QSizePolicy.Fixed,
                                qtw.QSizePolicy.Preferred)
            combo.setMinimumWidth(self.aB.width())
        
        allWidgs.extend([self.bulkGroup, self.surfGroup])
        
        # Button to input extra symmetry operations for bulk
        self.bulk_3d_sym = qtw.QPushButton('Extra bulk\nsymmetry')
        self.bulk_3d_sym.setFont(gl.AllGUIFonts().buttonFont)
        self.bulk_3d_sym.setSizePolicy(qtw.QSizePolicy.Fixed,
                                           qtw.QSizePolicy.Preferred)
        self.bulk_3d_sym.setMinimumWidth(self.aB.width())
        allWidgs.append(self.bulk_3d_sym)
        
        
        # Text and button for reduction to highest symmetry                     # TODO: this does not really work nicely. Also, its layout is not inserted (commented out below)
        self.highSymTxt = qtw.QLabel('The lattices could have higher symmetry')
        self.highSymBut = qtw.QPushButton('Reduce to higher symmetry')
        for widg in [self.highSymTxt, self.highSymBut]:
            widg.setFont(smallText)
            widg.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Preferred)
        self.highSymBut.setMinimumWidth(self.highSymBut.width()*0.4)
        self.highSymTxt.setMinimumWidth(self.highSymBut.width()*0.4)
        
        # Superstructure input: dropdown and 2x2 matrix
        superstructLab = qtw.QLabel('Reconstruction periodicity')
        superstructLab.setFont(labFont)
        woodsLab = qtw.QLabel("Wood's:")
        woodsLab.setFont(smallText)
        woodsLab.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Preferred)
        self.woods = qtw.QComboBox()
        self.woods.setEditable(True)
        self.woods.setInsertPolicy(qtw.QComboBox.InsertAlphabetically)
        self.woods.setFont(smallText)
        self.woods.setSizePolicy(qtw.QSizePolicy.Fixed,
                                 qtw.QSizePolicy.Preferred)
        self.woods.setSizeAdjustPolicy(qtw.QComboBox.AdjustToContents)
        self.woods.setMinimumContentsLength(12)
        
        superlatticeLab = qtw.QLabel('M = ')
        superlatticeLab.setFont(labFont)
        parFont = qtg.QFont(labFont)
        parFont.setPointSize(30)
        superlatticeLeftP = qtw.QLabel('(')
        superlatticeLeftP.setFont(parFont)
        superlatticeRightP = qtw.QLabel(')')
        superlatticeRightP.setFont(parFont)
        self.superlattice = np.array([[qtw.QLineEdit(), qtw.QLineEdit()],
                                      [qtw.QLineEdit(), qtw.QLineEdit()]])
        intValidator = qtg.QIntValidator()
        for mij in self.superlattice.ravel():
            mij.setFont(labFont)
            mij.setMaximumWidth(35)
            mij.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Preferred)
            mij.setValidator(intValidator)
        
        allWidgs.extend([superstructLab, self.woods, superlatticeLab,
                         *self.superlattice.ravel()])
        
        # maximum LEED energy
        emaxLab = qtw.QLabel('Max. Energy:')
        self.eMax = qtw.QLineEdit()
        for ctrl in [emaxLab, self.eMax]:
            ctrl.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Preferred)
            ctrl.setFont(labFont)
        self.eMax.setMaximumWidth(70)
        allWidgs.extend([emaxLab,self.eMax])
        
        # 'Done' and 'Cancel' buttons
        self.doneBut = qtw.QPushButton('&Done')
        self.cancelBut = qtw.QPushButton('&Cancel')
        for but in [self.doneBut, self.cancelBut]:
            but.setFont(gl.AllGUIFonts().buttonFont)
            but.setSizePolicy(qtw.QSizePolicy.Fixed, qtw.QSizePolicy.Preferred)
        allWidgs.extend([self.doneBut, self.cancelBut])
        
        # status bar
        self.sBar = qtw.QStatusBar()
        self.sBar.setSizeGripEnabled(False)
        
        #------- Make sure all styles are up to date ------#
        [widg.ensurePolished() for widg in allWidgs]
        
        #------- Place controls in layouts -------#
        lattsLay = qtw.QGridLayout()
        lattsLay.setSpacing(5)
        lattsLay.setContentsMargins(5, 5, 5, 5)  # left, top, right, bottom
        lattsLay.setSizeConstraint(qtw.QLayout.SetFixedSize)
        lattsLay.addWidget(lattInputLabel, 0, 1, 1, 2)
        lattsLay.addWidget(self.bulkInput, 1, 1, 1, 1)
        lattsLay.addWidget(self.surfInput, 1, 2, 1, 1)
        lattsLay.addWidget(latShapeLabel, 2, 0, 1, 1)
        lattsLay.addWidget(self.bulkShape, 2, 1, 1, 1)
        lattsLay.addWidget(self.surfShape, 2, 2, 1, 1)
        lattsLay.addWidget(aLab, 3, 0, 1, 1)
        lattsLay.addWidget(self.aB, 3, 1, 1, 1)
        lattsLay.addWidget(self.aS, 3, 2, 1, 1)
        lattsLay.addWidget(bLab, 4, 0, 1, 1)
        lattsLay.addWidget(self.bB, 4, 1, 1, 1)
        lattsLay.addWidget(self.bS, 4, 2, 1, 1)
        lattsLay.addWidget(alphaLab, 5, 0, 1, 1)
        lattsLay.addWidget(self.alphaB, 5, 1, 1, 1)
        lattsLay.addWidget(self.alphaS, 5, 2, 1, 1)
        lattsLay.addWidget(groupLab, 6, 0, 1, 1)
        lattsLay.addWidget(self.bulkGroup, 6, 1, 1, 1)
        lattsLay.addWidget(self.surfGroup, 6, 2, 1, 1)
        lattsLay.addWidget(self.bulk_3d_sym, 7, 1, 1, 1)
        
        hSymLay = qtw.QVBoxLayout()
        hSymLay.setSpacing(4)
        hSymLay.setContentsMargins(0, 10, 0, 10)
        hSymLay.addWidget(self.highSymTxt)
        hSymLay.addWidget(self.highSymBut)
        hSymLay.addStretch(1)
        
        lattsLay.addLayout(hSymLay, 8, 0, 1, 3)
        
        for lab in [lattInputLabel, self.bulkInput, self.surfInput]:
            lattsLay.setAlignment(lab, qtc.Qt.AlignCenter)
        for lab in [*lattLabels, groupLab]:
            lattsLay.setAlignment(lab, qtc.Qt.AlignRight | qtc.Qt.AlignVCenter)
        
        matrixLay = qtw.QGridLayout()
        matrixLay.setSpacing(2)
        matrixLay.setContentsMargins(0, 0, 0, 0)
        [matrixLay.addWidget(self.superlattice[i,j], i, j+1)
         for i in range(2) for j in range(2)]
        outerMatrixLay = qtw.QHBoxLayout()
        outerMatrixLay.addWidget(superlatticeLab)
        outerMatrixLay.addWidget(superlatticeLeftP)
        outerMatrixLay.addLayout(matrixLay)
        outerMatrixLay.addWidget(superlatticeRightP)
        for txt in [superlatticeLab, superlatticeLeftP, superlatticeRightP]:
            outerMatrixLay.setAlignment(txt, qtc.Qt.AlignCenter)
        outerMatrixLay.addStretch(1)
        
        woodsLay = qtw.QHBoxLayout()
        woodsLay.addWidget(woodsLab)
        woodsLay.addWidget(self.woods)
        woodsLay.addStretch(1)
        
        superstructLay = qtw.QVBoxLayout()
        superstructLay.addWidget(superstructLab)
        superstructLay.addLayout(woodsLay)
        superstructLay.addLayout(outerMatrixLay)
        superstructLay.addStretch(1)
        
        eLay = qtw.QHBoxLayout()
        eLay.setSpacing(4)
        eLay.setContentsMargins(5, 5, 5, 5)
        eLay.addWidget(emaxLab)
        eLay.addWidget(self.eMax)
        eLay.addStretch(1)
        eLay.setAlignment(emaxLab, qtc.Qt.AlignRight | qtc.Qt.AlignVCenter)
        
        botButsLay = qtw.QHBoxLayout()
        botButsLay.addWidget(self.sBar)
        botButsLay.addWidget(self.doneBut)
        botButsLay.addWidget(self.cancelBut)
        
        # Place the layouts in parent layouts, and assemble 
        # the layout of the QDialog
        latticeInputsLay = qtw.QVBoxLayout()
        latticeInputsLay.addLayout(lattsLay)
        latticeInputsLay.addStretch(1)
        
        diagLay = qtw.QGridLayout()
        diagLay.addLayout(latticeInputsLay, 0, 0, 2, 1)
        diagLay.addLayout(superstructLay, 0, 2, 1, 1)
        #diagLay.addLayout(hSymLay, 1, 2, 1, 1)
        diagLay.addLayout(eLay, 1, 2, 1, 1)
        diagLay.addLayout(botButsLay, 2, 0, 1, 3)
        diagLay.setColumnMinimumWidth(1, 15)
        diagLay.setAlignment(eLay, qtc.Qt.AlignLeft | qtc.Qt.AlignVCenter)
        diagLay.setSizeConstraint(qtw.QLayout.SetFixedSize)
        
        self.setLayout(diagLay)
        
        #------ Connect signals ------#
        self.connectSignals()
    
    def connectSignals(self):
        self.bulkInput.toggled.connect(self.bulkToggled)
        for shape in [self.bulkShape, self.surfShape]:
            shape.activated.connect(self.lattShapeChanged)
        for group in [self.bulkGroup, self.surfGroup]:
            group.activated.connect(self.groupChanged)
        self.woods.activated.connect(self.woodsSelected)
        self.woods.lineEdit().editingFinished.connect(self.woodsSelected)
        for tbox in [*self.bulkParams, *self.surfParams]:
            tbox.textEdited.connect(self.lattParamsChanged)
            tbox.editingFinished.connect(self.updateLatticeParameters)
        for el in self.superlattice.ravel():
            el.textEdited.connect(self.matrixChanged)
        for but in [self.doneBut, self.cancelBut]:
            but.clicked.connect(self.onButtonPressed)
            # The next two lines prevent the dialog to be closed automatically
            # by firing one of the buttons if the user presses enter/return
            # while focus is on widgets other than the buttons. The
            # reimplementation of keyPressEvent takes care of the behavior on 
            # pressing enter/return
            but.setAutoDefault(False)
            but.setDefault(False)
        self.bulk_3d_sym.clicked.connect(self.on_bulk_3d_pressed)
        self.highSymBut.clicked.connect(self.onReducePressed)
    
    # The next 7 functions are the handler of control changes
    
    def bulkToggled(self, checked):
        bulk = [*self.bulkParams, self.bulkShape]
        surf = [*self.surfParams, self.surfShape]
        if checked:
            enable = bulk
            disable = surf
        else:
            enable = surf
            disable = bulk
        [box.setEnabled(True) for box in enable]
        [box.setEnabled(False) for box in disable]
        
        self.updateLatticeRestrictions()
    
    def lattShapeChanged(self, signal=None):
        #--- Set the appropriate relations depending on the selection ---#
        if self.bulkInput.isChecked():
            shape = self.bulkShape
            params = self.bulkParams  #[a, b, alpha]
        else:
            shape = self.surfShape
            params = self.surfParams
        
        shape = shape.currentText()
        (a, b, alpha) = tuple(params)
        
        if shape in ['Square', 'Rhombic', 'Hexagonal']:
            b.setEnabled(False)
            b.setText(a.text())
        else:
            b.setEnabled(True)
        
        if shape in ['Square', 'Rectangular', 'Hexagonal']:
            alpha.setEnabled(False)
            if shape in ['Square', 'Rectangular']:
                alpha.setText('90' + degrees)
            else:
                alpha.setText('120' + degrees)
        else:
            alpha.setEnabled(True)
        
        #--- react on the changes of lattice parameters ---#
        self.lattParamsChanged()
        
        #--- update the lattice shape of the other ---#
        if self.bulkInput.isChecked():
            other = 'surf'
            self.surfShape.setCurrentText(self.surfLatt.cell_shape)
        else:
            other = 'bulk'
            self.bulkShape.setCurrentText(self.bulkLatt.cell_shape)
        
        #--- and update the dependent controls ---#
        if signal is not None:
            # method has been called as a result of a user action
            other = 'both'
        self.updateLatticeParameters(other)
        self.updateWoodsList()
        self.updateGroups()
    
    def lattParamsChanged(self, newText=None):
        if newText is not None:
            # The function is called as a result of a user interaction
            #
            # check that the input is valid (the user could be typing text)
            # ADD MATH PARSER HERE
            reNumber = re.compile(r'^\s*\d+([.]\d+)?\s*[\u212b\u00b0]?$')
            number = reNumber.match(newText)
            if number is not None:  # input is a number (with unit)
                self.updateLatticeRestrictions()
        
        if self.bulkInput.isChecked():
            type = 'bulk'
        else:
            type = 'surf'
        
        basis = self.getBasis(type)
        if basis is not None:  # edit value is OK
            self.updateLattices(basis)
            self.updateSymmReduction()
    
    def matrixChanged(self, newEl=None):
        if newEl is None:
            newEl = self.superlattice[0, 0].text()
        #--- check that the input is an integer ---#
        v = self.superlattice[0, 0].validator()
        valid = v.validate(newEl, 1)[0] == qtg.QValidator.Acceptable
        
        if valid: # update
            texts = [el.text() for el in self.superlattice.ravel()]
            m = np.array([int(text) for text in texts]).reshape(2, 2)
            #--- check that the input gives a non-singular matrix ---#
            if np.linalg.det(m) != 0:
                self.supMatrix = m.reshape(2, 2)
            
                self.updateLattices()
                self.updateLatticeParameters()
                self.updateLatticeRestrictions()
                self.updateWoods()
                self.sBar.clearMessage()
            else:
                self.sBar.showMessage('Matrix is singular!', 2000)
    
    def woodsSelected(self, woodsIdx=None):
        if woodsIdx is None:
            # the call to the function came from the editingFinished of the
            # QLineEdit
            woods = self.woods.lineEdit().text()
        else:
            # the call came by a user click on a QComboBox item
            woods = self.woods.itemText(woodsIdx)
        woods = self.fixWoods(woods)
        if woods is not None:
            m = self.woodL.woodsToMatrix(woods, self.bulkLatt.basis)
            if m is None:  # woods gives an incommensurate lattice
                self.sBar.showMessage(woods + ' invalid: Lattice is '
                                      + 'incommensurate.', 7000)
                self.woods.removeItem(self.woods.currentIndex())
                self.updateWoods()
            else:
                self.updateSuperlattice(m)
                self.matrixChanged()
        else:
            self.sBar.showMessage("Invalid Wood's syntax.", 1000)
    
    def groupChanged(self, newGroup=None):
        if newGroup is not None:
            self.updateLatticeGroups()
    
    def onButtonPressed(self, checked):
        btn = self.sender()
        if btn == self.doneBut:
            # Automatically make lattices high symmetry before exiting          # TODO: this screws up for the extra bulk operations. The high-symmetry button should rather appear after edits.
            # self.highSymBut.click()
            params = self.packParameters()
            if params is not None:
                self.dialogEdited.emit(params)
                self.accept()
        elif btn == self.cancelBut:
            self.dialogEdited.emit(self.oldParams)
            self.reject()
    
    def onReducePressed(self, checked):
        tBulk = self.bulkLatt.make_high_symmetry()
        tSurf = self.surfLatt.make_high_symmetry()
        m = np.dot(np.dot(tSurf, self.supMatrix), np.linalg.inv(tBulk))
        #  now make entries integer
        m = m.round().astype(int)
        
        self.updateLatticeParameters()
        self.updateSuperlattice(m)
        self.updateLatticeRestrictions()
        self.updateWoodsList()
        self.updateGroups()
    
    def on_bulk_3d_pressed(self, checked):
        self.bulk_3d_sym_dialog.update_operations(self.bulkLatt)
        if self.bulk_3d_sym_dialog.exec() == qtw.QDialog.Accepted:
            operations = self.bulk_3d_sym_dialog.extra_operations()
            self.bulkLatt.group.screws_glides = (operations,
                                                 self.bulkLatt.cell_shape)
    
    # The next 8 function are explicitly called to handle changes of controls
    
    def updateLattices(self, basis = None):
        #------- Update lattices and their lattice parameters -------#
        # If bulk is selected, the surface parameters will be changed
        # at constant bulk parameters. Otherwise the other way around.
        
        if self.bulkInput.isChecked():
            if basis is None:
                basis = self.bulkLatt.basis
            else:
                bulk = self.bulkLatt
                bulk.basis = basis
            latt = self.surfLatt
            m = self.supMatrix
        else:
            if basis is None:
                basis = self.surfLatt.basis
            else:
                surf = self.surfLatt
                surf.basis = basis
            latt = self.bulkLatt
            m = np.linalg.inv(self.supMatrix)
        
        latt.basis = np.dot(m, basis)
        
    def updateLatticeGroups(self):
        bGroup = self.bulkGroup.currentText()
        sGroup = self.surfGroup.currentText()
        allGroups = self.bulkLatt.group.allGroups
        for (latt, group) in zip([self.bulkLatt, self.surfLatt],
                                 [bGroup, sGroup]):
            if group in allGroups:
                bulk_3d = latt.group.screws_glides
                latt.group = group
                latt.group.screws_glides = (bulk_3d, latt.cell_shape)
        
        # Also, update the options for extra bulk operations
        self.bulk_3d_sym_dialog.update_operations(self.bulkLatt)
        # and deactivate button in case there is nothing to add
        self.bulk_3d_sym.setEnabled(self.bulk_3d_sym_dialog.n_extra_ops)
    
    def updateLatticeParameters(self, which='both'):
        if which == 'bulk':
            params = [self.bulkParams]
            latts = [self.bulkLatt]
        elif which == 'surf':
            params = [self.surfParams]
            latts = [self.surfLatt]
        else:
            params = [self.bulkParams, self.surfParams]
            latts = [self.bulkLatt, self.surfLatt]
        for (ctrls,latt) in zip(params, latts):
            a, b, alpha = latt.lattice_parameters
            ctrls[0].setText('{:.4f}{}'.format(a, angstrom))
            ctrls[1].setText('{:.4f}{}'.format(b, angstrom))
            ctrls[2].setText('{:.1f}{}'.format(alpha, degrees))
    
    def updateSuperlattice(self, m=None):
        if m is None:
            m = self.supMatrix
        m = np.array(m)
        for (el, mij) in zip(self.superlattice.ravel(), m.ravel()):
            el.setText(str(int(mij)))
        self.supMatrix = m
    
    def updateLatticeRestrictions(self):
        self.lattShapeChanged()
    
    def updateWoodsList(self):
        # loads the example list of woods notations
        # for the selected bulk shape
        shape = self.bulkLatt.cell_shape
        self.woods.clear()
        self.woods.addItems(sorted(self.woodL.examples[shape]))
        self.updateWoods()
    
    def updateWoods(self):
        woods = gl.Woods().matrixToWoods(self.supMatrix, self.bulkLatt.basis)
        if woods is None:
            woods = 'None'
        idx = self.woods.findText(woods, flags = qtc.Qt.MatchExactly)
        if idx < 0:  # item is not already present
            self.woods.addItem(woods)
            self.woodL.examples[self.bulkLatt.cell_shape] |= {woods}
        self.woods.setCurrentText(woods)
    
    def updateGroups(self):
        for (ctrl, latt) in zip([self.bulkGroup, self.surfGroup],
                                [self.bulkLatt, self.surfLatt]):
            shape = latt.cell_shape
            group = latt.group
            groups = group.groupsForShape[shape]
            ctrl.clear()
            ctrl.addItems(groups)
            ctrl.setCurrentText(group.group)
        # update the groups in the lattices, in case the 
        # current text and the group do not match
        self.updateLatticeGroups()
    
    def updateSymmReduction(self):
        tBulk = self.bulkLatt.high_symm_transform()
        tSurf = self.surfLatt.high_symm_transform()
        id = [[1, 0], [0, 1]]
        isHighSym = np.array_equal(tBulk, id) and np.array_equal(tSurf, id)
        for widg in [self.highSymBut, self.highSymTxt]:
            if isHighSym:
                widg.hide()
            else:
                widg.show()
    
    # and the next 4 functions are utilities
    
    def packParameters(self):
        reUnits = re.compile(r'(\d+(\.\d+)?)\s*[\u212b\u00b0eV]*') 
        # angstrom, degrees, eV
        eMax = reUnits.match(self.eMax.text())
        if eMax is None:
            return None
        eMax = float(eMax.group(1))
        
        params = {'eMax': eMax,
                  'SUPERLATTICE': self.supMatrix,
                  'surfBasis': self.surfLatt.basis,
                  'surfGroup': self.surfLatt.group,
                  'bulkGroup': self.bulkLatt.group,
                  'bulk3Dsym': self.bulkLatt.group.screws_glides}
        
        return params
    
    def getBasis(self, which = 'bulk'):
        reUnits = re.compile(r'(\d+(\.\d+)?)\s*[\u212b\u00b0]?')
        if which == 'bulk':
            params = [self.aB.text(), self.bB.text(), self.alphaB.text()]
        else:
            params = [self.aS.text(), self.bS.text(), self.alphaS.text()]
        params = [reUnits.match(param) for param in params]
        if None in params:
            return None
        params = [param.group(1) for param in params]
        
        for param in params:
            try:
                float(param)
            except:
                return None
        
        # All entries are floats
        (a, b, alpha) = (float(param) for param in params)
        alpha = np.radians(alpha)
        if a > 0 and b > 0:
            return np.array([[a, 0],[b*np.cos(alpha), b*np.sin(alpha)]])
        return None
    
    def fixWoods(self, woods):
        if woods is None:
            return None
        
        reWoods = re.compile(
            r'''
            ^(?P<prefix> ((?![rR])[\sa-zA-Z])*)     # * Multi-letter
                                                    #   prefix -> will turn 
                                                    #   into p or c
            \(?                                     # * Optional open 
                                                    #   parenthesis
            (?P<g1int>\d+)?[*]?                     # * Dir1, integer part,
                                                    #   optional (+ optional 
                                                    #   asterisk for 
                                                    #   multiplication)
            (?:[sqSQ]{0,2}[rR\u221a][tT]?\(?(?P<g1rt>\d+)\)?)?
                                                    # * Dir1, sqrt (and 
                                                    #   similar) part
            [x\u00d7]                               # * One times character
            (?P<g2int>\d+)?[*]?                     # * Dir2, integer part,
                                                    #   optional (+ optional 
                                                    #   asterisk for 
                                                    #   multiplication)
            (?:[sqSQ]{0,2}[rR\u221a][tT]?\(?(?P<g2rt>\d+)\)?)?
                                                    # * Dir2, sqrt (and 
                                                    #   similar) part
            \)?                                     # * Optional closing
                                                    #   parenthesis
            ([rR](?P<alpha>\d+(\.\d+)?)\u00b0?)?$   # * rotation block
            ''', re.VERBOSE)
        
        m = reWoods.match(woods)
        if m is None:
            return None
        
        groups = m.groupdict()
        if 'c' in groups['prefix']:  # force prefix to be p or c
            groups['prefix'] = 'c'
        else:
            groups['prefix'] = 'p'
        
        for key in ['g1int' ,'g2int']:
            # replace the empty integer parts with 1
            if groups[key] is None:
                groups[key] = 1
            else:
                groups[key] = float(groups[key])
        
        toFormat = []
        dirs = [(groups['g1int'], groups['g1rt']),
                (groups['g2int'], groups['g2rt'])]
        for (integral, rt) in dirs:
            # Now combine integer and square root parts
            if rt is not None:
                sq = float(rt)
                (intSq, rad) = self.woodL.squareToProdOfSquares(sq)
                integral *= np.round(np.sqrt(intSq))
            else:
                rad = 1
            
            dirStr = ''
            if rad > 1:
                dirStr = '\u221a' + str(int(rad))
            if not dirStr:
                dirStr = str(int(integral))
            else:
                if integral > 1:
                    dirStr = str(int(integral)) + dirStr
            toFormat.append(dirStr)
        
        woodsFixed = '{}({})'.format(groups['prefix'],'\u00d7'.join(toFormat))
        
        #finally handle the angle
        if groups['alpha'] is not None:
            alpha = float(groups['alpha'])
            cA = np.abs(np.cos(np.radians(alpha)))
            if cA > 1e-3 and 1-cA > 1e-3:
                # angle is neither 0, 90, nor 180
                woodsFixed += 'R{:.1f}'.format(alpha) + degrees
        
        return woodsFixed
    
    def initControlValues(self):
        #---- Explicitly set some values ----#
        self.bulkInput.setChecked(True)
        # always set bulk checked at session startup
        # -> might remember choice later
        self.bulkShape.setCurrentText(self.bulkLatt.cell_shape)
        self.surfShape.setCurrentText(self.surfLatt.cell_shape)
        if 'eMax' in self.oldParams:
            e = self.oldParams['eMax']
        else:
            e = 700
        self.eMax.setText(str(e) + ' eV')
        for widg in [self.highSymTxt, self.highSymBut]:
            widg.hide()
        
        #---- Call some updating functions ----#
        self.updateLatticeParameters()
        self.updateSuperlattice()
        self.updateLatticeRestrictions()
        self.updateWoodsList()
        self.updateGroups()
    
    # Re-implement keyPressEvent so that pressing enter or return when Done or
    # Cancel have keyboard focus fires them
    
    def keyPressEvent(self, event):
        if event.key() in [qtc.Qt.Key_Enter or qtc.Qt.Key_Return]:
            focusWidg = self.focusWidget()
            if isinstance(focusWidg, qtw.QLineEdit):
                return
            elif (isinstance(focusWidg, qtw.QComboBox)
                  and focusWidg.isEditable()):
                return
            elif focusWidg == self.cancelBut:
                self.cancelBut.click()
                return
            self.doneBut.click()
            return
        #super(NewFileDialog, self).keyPressEvent(event)
        super().keyPressEvent(event)



"""Module r_error of viperleed.calc.classes.

Defines data structures and functions for storing and manipulating
errors curves.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-01-12'
__license__ = 'GPLv3+'

import copy
import logging

import numpy as np

from viperleed.calc.lib.base import range_to_str


logger = logging.getLogger(__name__)


class R_Error():
    """Data structure for storing errors after they have been calculated.

    Stores the displacements that led to the R-factors even if they are
    modified later. Displacements are stored for one atom only, the other are
    linked.
    """

    def __init__(self, r_type, atoms, mode, rfacs, disp_label, lin_disp,
                 v0i=None, energy_range=None):
        self.r_type = r_type # Type of R-factor (int). Only Pendry R-factor (r_type=1) can give statistical error estimates.
        self.atoms = atoms  # atoms that have been varied together
        self.mode = mode    # vib, geo, or occ
        self.rfacs = rfacs  # the r-factors from the variations
        self.var_r = None
        self.displacements = []   # displacements of atoms[0]
        self.disp_label = disp_label
        self.lin_disp = lin_disp  # linearized displacement
        self.main_element = ""    # element occupation displayed in output
        self.elem_occ = None
        d = {}
        at = atoms[0]   # store displacements for this one; main element
        if mode == "occ":
            if atoms[0].el in atoms[0].disp_occ:
                self.main_element = atoms[0].el  # main site element
            else:
                # element with highest occupation in refcalc
                self.main_element = max(atoms[0].site.occ,
                                        key=lambda k: atoms[0].site.occ[k])
            self.displacements = copy.deepcopy(atoms[0].disp_occ[
                                                        self.main_element])
            self.lin_disp = copy.deepcopy(self.displacements)
            self.elem_occ = copy.deepcopy(atoms[0].disp_occ)
            vac_occ = np.ones_like(self.lin_disp)
            for el_occ in self.elem_occ.values():
                vac_occ -= el_occ
            if np.sum(vac_occ) > 1e-4:  # If vacancies, put them into output CSV
                self.elem_occ["Vac"] = vac_occ
        else:
            if mode == "geo":
                d = atoms[0].disp_geo
            elif mode == "vib":
                d = atoms[0].disp_vib
        if len(d) > 0:
            if at.el in d:
                k = at.el
            else:
                k = "all"
            self.displacements = copy.deepcopy(d[k])

        if v0i and energy_range and r_type==1:
            self.calc_var_r(v0i, energy_range)


    def calc_var_r(self, v0i, energy_range):
        r"""Calculates the variance of the Pendry R-factor.

        The variance of the Pendry R-factor can be estimated from the
        imaginary part of the inner potential and the measurement energy
        range as:

        .. math::
            \mathrm{var}(R_p) = \sqrt{ \frac{8V_{0\mathrm{i}}}{E_{\mathrm{range}}}}R_{\mathrm{min}}

        Parameters
        ----------
        v0i : float
            Imaginary part of the inner potential (eV).
        energy_range : float
            Energy range in eV.

        Raises
        ------
        ValueError:
            If v0i or energy_range are not float values.
        ValueError:
            If v0i or energy_range are negative. This would be
            unphysical.
        """
        if self.r_type != 1:
            raise ValueError("Var(R) can only be calculated when using the "
                             "Pendry R-factor.")
        if not isinstance(v0i, float) and isinstance(energy_range, float):
            raise ValueError("v0i and energy_range must be float values.")
        if v0i<= 0 or energy_range <= 0:
            raise ValueError("v0i and energy range must be positive ")
        if not self.rfacs:
            raise ValueError("Cannot calculate variance of R-factor without "
                             "stored R-factors.")
        self.var_r = np.sqrt(8*np.abs(v0i) / energy_range) * self.get_r_min


    @property
    def get_error_estimates(self):
        """Decide if statistical error estimates can be given and
        calculates them.

        Returns
        -------
        tuple of float or None
            A tuple containing upper and lower error bounds, es
            estimated from the variance of R_P, if errors can be
            estimated. If either bound can not be estimates (e.g. no
            intersection with line R_min + var(R)), the element is None.
            If neither bound can be estimated, both values are None.

        Raises
        ------
        ValueError
            If self.lin_disp is not initialized.
        """

        # error estimates are only available for Pendry R-factor
        if self.r_type != 1:  #TODO: when we introduce a new R-factor, make sure to update this.
            return (None, None)
        r_min = self.get_r_min
        if self.var_r is None:
            logger.warning("Cannot calculate statistical errors for "
                           f'atoms {range_to_str([at.num for at in self.atoms])}')
            return (None, None)

        if self.lin_disp is None:
            raise ValueError("Linear displacements not initialized.")

        # more than 4 points
        if len(self.rfacs) < 3 or not self.var_r:
            return (None, None)

        # minimum R factor of this error is not necessarily the same as
        # the overall minimum R factor
        err_min_R = min(self.rfacs)

        # TODO change!!!!!
        if err_min_R > r_min + self.var_r:
            return (None, None)

        # treat upper and lower half of error curve
        err_min_R_id = self.rfacs.index(err_min_R)
        best_param = self.lin_disp[self.rfacs.index(err_min_R)]
        left_rfacs = self.rfacs[:err_min_R_id+1]
        left_lin_disp = self.lin_disp[:err_min_R_id+1]
        right_rfacs = self.rfacs[err_min_R_id:]
        right_lin_disp = self.lin_disp[err_min_R_id:]

        error_estimates = []
        for rfacs, lin_disp in ((left_rfacs, left_lin_disp),
                                (right_rfacs, right_lin_disp)):
            # find zero crossing
            offset_rfacs = rfacs - (r_min + self.var_r)
            if (len(rfacs) < 3 or                                         # too few points
                get_n_zero_crossings(offset_rfacs) != 1): # unclear crossing of line
                error_estimates.append(None)
            elif(max(rfacs) <= r_min + self.var_r):       # maximum R below r_min + var(R) line
                error_estimates.append(None)
            else:
                crossing_point = get_zero_crossing(lin_disp, offset_rfacs)
                delta = abs(best_param - crossing_point)
                error_estimates.append(delta)
        return tuple(error_estimates)

    @property
    def get_r_min(self):
        """Returns minimum R-factor of error.

        Returns
        -------
        float
            Minimum R-factor stored in self.rfacs.
        """
        return min(self.rfacs)

    @property
    def get_p_min(self):
        """Returns parameter value at minimum R-factor.

        Returns
        -------
        float
            parameter value at minium R-factor (self.rfacs).
        """
        return self.lin_disp[self.rfacs.index(self.get_r_min)]


def get_n_zero_crossings(x_arr):
    """Counts and returns the number of zero crossings in input.

    x_arr must be discrete numerical data. This function counts the
    number of times the zero line is crossed when going through the
    input values in order, e.g. x_arr=[-3, 1 , 1, -5, -2] would return
    2.

    Parameters
    ----------
    x_arr : list or ndarray of numbers
        Discrete numerical input data.

    Returns
    -------
    int
        Number of times the discrete input data crossed the zero line.
    """
    arr = np.array(x_arr)
    n_zero_crossings = int((1 - np.sign((arr[:-1]*arr[1:]))).sum()/2)
    return n_zero_crossings


def get_zero_crossing(x_arr, y_arr, eps=1e-6):
    """Computes and returns the x_value where the provided dataset crosses zero.

    x_arr and y_arr are sets of x and y values respectively. This
    function first checks in between which points the zero line is
    crossed. It then performs a linear interpolation between those two
    points and gives the interpolated x values, where the zero line is
    crossed. If two subsequent points are zero, the middle is returned.
    If the dataset crosses the zero line multiple times, only the first
    x value will be returned. See also get_n_zero_crossings().

    Parameters
    ----------
    x_arr : list or array of numbers.
        x_values of the data to be checked.
        Must in ascending order.
    y_arr : _type_
        _description_
    eps: float
        Numerical accuracy of comparison.

    Returns
    -------
    float or None
        The linearly interpolated x value of the zero crossing.
        None if no zero crossing is found.

    Raises
    ------
    ValueError
        if x_arr and y_arr have incompatible shapes.
    ValueError
        if the values of x_arr are not monotonically increasing.
    """

    if len(x_arr) != len(y_arr):
        raise ValueError("Different sized arrays supplied.")
    # make sure x_arr is monotonically increasing
    if not np.all(x_arr[1:] >= x_arr[:-1]):
        raise ValueError("Values in x_arr are not monotonically increasing")

    # find location of zero crossing
    for id in range(len(x_arr)-1):
        if y_arr[id]*y_arr[id+1] <= 0:
            R_jump = abs(y_arr[id+1]-y_arr[id])
            if R_jump < eps:  # avoid division by zero
                return (x_arr[id]+x_arr[id+1])/2
            return x_arr[id]-y_arr[id]*(x_arr[id+1]-x_arr[id])/(y_arr[id+1]-y_arr[id])
    # return None if no zero crossing found
    return None



"""Module for reading and writing Woods notation.

The functionality in this module used to be part of calc.lib.leedbase.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-06-07'
__license__ = 'GPLv3+'

import logging
import re

import numpy as np

from viperleed.calc.lib.base import angle, cosvec, parseMathSqrt

logger = logging.getLogger(__name__)

def readWoodsNotation(s, ucell):
    """Takes a string that should contain the transformation from the bulk to
    the surface unit cell in Wood notation, as well as a bulk unit cell (from
    which only the surface vectors are read). Returns a 2x2 transformation
    matrix."""
    p = re.compile(r'\s*(?P<type>[PCpc]*)\s*\(\s*(?P<g1>.+)\s*[xX]\s*'
                   + r'(?P<g2>.+)\s*\)\s*[rR]*\s*(?P<alpha>[\d.]*)')
    # this regular expression matches if (any amount of whitespace at any
    #                                                     point is ignored):
    #   - optional: first character is p or c (or P or C) (-> type)
    #   - then there is a '('
    #   - then whatever (at least one character), interpret later (-> g1)
    #   - then 'x' or 'X'
    #   - then whatever (at least one character), interpret later (-> g2)
    #   - then ')'
    #   - then (optional) 'r' or 'R'
    #   - then (optional) an integer or float number (-> alpha)
    m = p.match(s)
    if not m:
        logging.error('Could not read woods notation input '+s)
        return None
    if not m.group('type'):
        t = 'p'
    else:
        t = m.group('type').lower()
    if not m.group('alpha'):
        alpha = 0.0
    else:
        try:
            alpha = float(m.group('alpha'))
        except ValueError:
            logger.error('Could not read Woods notation angle: '
                         + m.group('alpha')+', setting angle to zero')
            alpha = 0.0
    alpha *= np.pi/180
    g1 = parseMathSqrt(m.group('g1'))
    g2 = parseMathSqrt(m.group('g2'))
    # get surface unit cell vectors from bulk unit cell (has surface
    #  periodicity!!):
    if alpha == 0.0 and t == 'p':
        mat = np.array([[g1, 0.], [0., g2]], dtype=float)
    else:
        r = [ucell[:2, 0], ucell[:2, 1]]
        # q = np.linalg.norm(r[1])/np.linalg.norm(r[0])
        # this would be to get from bulk vectors to surface, we have to reverse
        q = 1/(np.linalg.norm(r[1])/np.linalg.norm(r[0]))
        omega = abs(angle(r[0], r[1]))
        # this is always constant in Wood notation, no need to reverse.
        if t == 'p':
            # matrices from: Klaus Hermann; Crystallography and Surface
            #                               Structure (Second Edition, Wiley)
            mat = ((1/np.sin(omega))
                   * np.array([[g1*np.sin(omega-alpha),
                                g1*(1/q)*np.sin(alpha)],
                               [-g2*q*np.sin(alpha),
                                g2*np.sin(omega+alpha)]],
                              dtype=float))
        else:
            mat = ((1/(2*np.sin(omega)))
                   * np.array([[g1*np.sin(omega-alpha)-g2*q*np.sin(alpha),
                                g1*(1/q)*np.sin(alpha)+g2*np.sin(omega+alpha)],
                               [-g1*np.sin(omega-alpha)-g2*q*np.sin(alpha),
                                -g1*(1/q)*np.sin(alpha)+g2*np.sin(omega
                                                                  + alpha)]],
                              dtype=float))
    warn = False
    for i in range(0, 2):
        for j in range(0, 2):
            if abs(mat[i, j] - round(mat[i, j])) < 1e-4:
                mat[i, j] = round(mat[i, j])
            else:
                warn = True
    if warn:
        logger.warning("SUPERLATTICE matrix from Woods notation was "
                       "identified as:\n"+str(mat))
        logger.warning("SUPERLATTICE values do not round to "
                       "integer values. Check SUPERLATTICE parameter.")
    return mat

def writeWoodsNotation(ucell):
    """Takes a unit cell (as a (2x2) matrix) and attempts to write it in Woods
    Notation. Returns empty string if no Woods notation is found."""
    # !!! VERY INCOMPLETE, should at least detect simple c(a x b) cases
    # !!! Same functionality exists in guilib; replace at some point
    if ucell[1, 0] == 0 and ucell[0, 1] == 0:
        return("(" + str(int(ucell[0, 0])) + "x" + str(int(ucell[1, 1])) + ")")
    else:
        return ""



"""Module iorfactor of viperleed.calc.files.

Functions for reading and writing files relevant to the
R-factor calculation.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import logging
import os
import re

import fortranformat as ff
import numpy as np

try:
    import matplotlib
except ImportError:
    _CAN_PLOT = False
else:
    _CAN_PLOT = True
    matplotlib.rcParams.update({'figure.max_open_warning': 0})
    matplotlib.use('Agg')  # !!! check with Michele if this causes conflicts
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt
    import matplotlib.ticker as plticker
    plt.style.use('viperleed.calc')

from viperleed.calc.classes.rparams import EnergyRange
from viperleed.calc.files.beams import writeAUXEXPBEAMS
from viperleed.calc.files.ivplot import plot_iv
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.version import Version

logger = logging.getLogger(__name__)


# How many extra points to take at the boundaries to prevent wasting
# data for interpolation of theoretical beams? The FORTRAN code likes
# to have 2 intervals left and right. Notice that this value is
# specific of the FORTRAN code, which uses 3rd order interpolation.
# It is made into a const so we can change it in only one spot should
# I (@michele-riva) have gotten it wrong. It SHOULD NOT BE USED for
# other purposes.
_N_EXPAND_THEO = 2


class RfactorError(Exception):
    """Base exception for R-factor calculations."""


def readROUT(filename="ROUT"):
    """
    Reads the ROUT file.

    Parameters
    ----------
    filename : string, optional
        Pass if you want to read from a file other than 'ROUT'

    Returns
    -------
    tuple (r, r_int, r_frac), float v0rshift, list rfaclist
        r, r_int, r_frac: tuple of floats:
            average R-factor for all beams, integer-order beams and
            fractional order beams.
        v0rshift: inner potential shift corresponding to r, r_int, r_frac
        rfaclist: list of floats, r-factors per beam in order of experimental
            beams

    """
    try:
        with open(filename, 'r') as rf:
            lines = rf.readlines()
    except Exception:
        logger.error("Could not open " + filename + " file")
        raise
    line = ""
    i = 0
    # finds line at end of ROUT file that states best v0r and corresponding R-factor
    while "AVERAGE R-FACTOR =" not in line and i+1 < len(lines):
        i += 1
        line = lines[-i]
    if line == "":
        return (0, 0, 0), 0, []
    rfac = 0
    rfac_int = -1
    rfac_frac = -1
    v0rshift = 0
    # writes best V0r to v0rshift and best R-factor to rfac
    rgx = re.compile(r'.+SHIFT\s*(?P<shift>[-0-9.]+).+=\s*(?P<rfac>[-0-9.]+)')
    m = rgx.match(line)
    if m:
        try:
            rfac = float(m.group("rfac"))
        except ValueError:
            logger.error("Could not read R-factor from "+filename)
        try:
            v0rshift = float(m.group("shift"))
        except ValueError:
            logger.error("Could not read inner potential shift from "
                         + filename)
    else:
        return (0, 0, 0), 0, []
    # now read the R-factors per beam only at best V0r
    rfaclist = []
    for line in [li for li in lines if len(li) >= 70]:
        line = line.strip()
        if line.endswith("<---"):
            line = line[:-4]

        # rfactor.f reserves 5A4 i.e. 20 characters for the beam name
        values = line[19:].split()
        try:
            index = int(values[0])
            v0r = float(values[2])
            rav = float(values[-1])
        except (ValueError, IndexError):
            pass    # ignore line
        else:
            if v0r == v0rshift and index > 0:
                if index != len(rfaclist)+1:
                    logger.warning("Unexpected index mismatch in readROUT. "
                                   "Reading R-factors per beam will fail.")
                else:
                    rfaclist.append(rav)
            elif v0r == v0rshift and index == -1:
                if line.startswith("AV.-INT"):
                    rfac_int = rav
                elif line.startswith("AV.-FRAC"):
                    rfac_frac = rav
    return (rfac, rfac_int, rfac_frac), v0rshift, rfaclist


def readROUTSHORT(filename="ROUTSHORT"):
    """
    Reads the ROUTSHORT file. This is very minimalist and just contains one
    average R-factor per line.

    Parameters
    ----------
    filename : string, optional
        Pass if you want to read from a file other than 'ROUTSHORT'

    Returns
    -------
    rfaclist : list of floats
        r-factors from the ROUTSHORT file

    """
    rfaclist = []
    try:
        with open(filename, 'r') as rf:
            lines = rf.readlines()
    except Exception:
        logger.error("Could not open " + filename + " file")
        raise
    for line in lines:
        line = line.strip()
        if not line:
            continue
        try:
            rfaclist.append(float(line))
        except ValueError:
            logger.warning("Unexpected value in " + filename + " file, could "
                           "not transform to float: "+line)
    return rfaclist


def check_theobeams_energies(rpars, theobeams):
    """Complain if the energies in theobeams are inconsistent with rpars."""
    theo_grid = sorted_energies_from_beams(theobeams)
    theo_energies = EnergyRange.from_sorted_grid(theo_grid)
    if not theo_energies.contains(rpars.THEO_ENERGIES, ignore_step=True):
        raise ValueError(
            f'theobeams has theo_energies={theo_energies}, which does not '
            'contain all the energies in the THEO_ENERGIES parameter '
            f'({rpars.THEO_ENERGIES}). Did you load the wrong calculation?'
            )


def prepare_rfactor_energy_ranges(rpars, theobeams=None, for_error=False,
                                  n_expand=_N_EXPAND_THEO):
    f"""Return EnergyRange objects for experiment, theory and iv_shifts.

    Parameters
    ----------
    rpars : Rparams
        The object holding information about the current PARAMETERS.
        Attributes used: THEO_ENERGIES, expbeams, IV_SHIFT_RANGE
    theobeams : Sequence or None, optional
        Elements are Beam objects. They are the beams loaded from
        the calculation against which experimental data should be
        compared. If given, it is checked for consistency with the
        rpars.THEO_ENERGIES attribute. theobeams are considered to
        be consistent if rpars.THEO_ENERGIES is a subset of the
        energies in theobeams with the same step size. No check is
        performed if not given or None. Default is None.
    for_error : bool, optional
        Whether the R-factor calculation is meant to be used for
        error curves. Default is False.
    n_expand : int, optional
        How many steps should the theory range be expanded at both
        ends not to disregard available values that may be useful
        for interpolation. This is on top of the expansion that is
        considered due to shifting. Default is {_N_EXPAND_THEO}.

    Returns
    -------
    exp_range : EnergyRange
        The range of energies of the experimental beams.
    theo_range : TheoEnergies
        The relevant range of energies. This is derived from the
        rpars.THEO_ENERGIES value. It is appropriately expanded to
        account for relative shifting of the curves. Additionally,
        it is also expanded by n_expand steps to the left and right.
    iv_shift : IVShiftRange
        The range of shifts to be applied. If for_error, this is
        fixed to rpars.best_v0r value. Otherwise, it is the rpars
        attribute IV_SHIFT_RANGE. Notice that, in the latter case,
        the attribute is modified in place if it does not yet have
        a step defined. Its step is set to interp_step. Notice also
        that, as a consequence of this change, rpars.IV_SHIFT_RANGE
        may be expanded so that its bounds are integer multiples of
        the step.
    interp_step : float
        The energy step to be used for interpolation. Notice that
        this is ALWAYS taken from rpars.IV_SHIFT_RANGE.

    Raises
    ------
    ValueError
        If theobeams was given, but it is inconsistent with rpars.
    RfactorError
        If rpars.THEO_ENERGIES has any undefined attribute, if there
        are no beams in rpars.expbeams, or if rpars.IV_SHIFT_RANGE
        has any of its bounds undefined.
    RfactorError
        When for_error and either (i) no rpars.best_v0r is present,
        or (ii) the rpars.best_v0r value is inconsistent with the
        interpolation step, or (iii) rpars.IV_SHIFT_RANGE has no
        step.
    """
    if theobeams is not None:
        check_theobeams_energies(rpars, theobeams)  # May ValueError
    if not rpars.THEO_ENERGIES.defined:
        raise RfactorError('Cannot run an R-factor calculation before '
                           f'rpars.THEO_ENERGIES={rpars.THEO_ENERGIES} '
                           'has stop, start, and step defined')
    if not rpars.expbeams:
        raise RfactorError('No experimental beams loaded')

    # Prepare experimental range
    exp_grid = sorted_energies_from_beams(rpars.expbeams)
    exp_range = EnergyRange.from_sorted_grid(exp_grid)

    # Now handle iv_shift. Notice that _prepare_iv_shift_range always
    # sets the step to the one of IV_SHIFT_RANGE to ensure consistency
    iv_shift = _prepare_iv_shift_range(rpars, exp_range, for_error)

    small_step = min(exp_range.step, rpars.THEO_ENERGIES.step)
    if not iv_shift.is_fixed and iv_shift.step > small_step:
        # Warn only if we're not fixed. If we are, it may just be
        # that the user gave us this specific step value because
        # it is exactly an integer multiple of the fixed value
        # they desired
        logger.warning(f'Using interpolation step {iv_shift.step} eV, which '
                       'is coarser than the smallest step between experiment '
                       f'and theory ({small_step} eV). Consider using '
                       f'{small_step} or not defining a step for the '
                       'IV_SHIFT_RANGE parameter')

    # Finally, the relevant range of theory energies
    theo_range = _find_common_theory_experiment_range(rpars.THEO_ENERGIES,
                                                      exp_range, iv_shift,
                                                      n_expand)
    return exp_range, theo_range, iv_shift, iv_shift.step


def _prepare_iv_shift_range(rpars, experiment, for_error):
    """Return an iv_shift appropriate for an R-factor calculation."""
    if for_error and rpars.best_v0r is None:
        raise RfactorError(
            'Cannot run an error calculation if rpars.best_v0r is '
            'undefined. Did you forget to run a normal R-factor first?'
            )
    if not rpars.IV_SHIFT_RANGE.has_bounds:
        raise RfactorError('Cannot run an R-factor calculation if '
                           f'rpars.IV_SHIFT_RANGE={rpars.IV_SHIFT_RANGE} '
                           'misses one of its bounds')

    # Set up the step. This also ensures that the bounds are consistent
    if not for_error:
        # Notice that here we purposely do not take as step
        # min(step, IV_SHIFT_RANGE.step). This would otherwise
        # require us to potentially force a modification of the
        # step of IV_SHIFT_RANGE and, likely, its bounds (because
        # they are integer multiples). This seems like too much
        # fiddling with the user input. See also discussion at
        # viperleed/commit/d4626116f11fb0bf9bef6c228413047a7207d441
        step = min(rpars.THEO_ENERGIES.step, experiment.step)
        try:
            rpars.IV_SHIFT_RANGE.set_undefined_step(step)
        except RuntimeError:  # Was fixed and we're trying to unfix it
            start = rpars.IV_SHIFT_RANGE.min
            raise RfactorError(
                f'Cannot fix IV_SHIFT_RANGE to {start}. The automatic step '
                f'from experiment and theory ({step} eV) is inappropriate: '
                f'{start} must be an integer  multiple of step={step}. '
                'Provide an explicit step by setting IV_SHIFT_RANGE'
                ) from None
        return rpars.IV_SHIFT_RANGE

    # In the for_error case, always make sure to
    # use the step stored in rpars.IV_SHIFT_RANGE
    if not rpars.IV_SHIFT_RANGE.has_step:
        raise RfactorError(
            'Cannot run an error calculation if rpars.IV_SHIFT_RANGE has '
            'no step defined. Did you forget to run a normal R-factor first?'
            )
    iv_shift = rpars.IV_SHIFT_RANGE.fixed(rpars.best_v0r)
    try:
        iv_shift.set_undefined_step(rpars.IV_SHIFT_RANGE.step)
    except RuntimeError as exc:
        # We have tried to unfix a fixed range: best_v0r
        # is for the wrong interpolation step
        raise RfactorError(f'rpars.best_v0r={rpars.best_v0r} is inconsistent '
                           f'with the interpolation step ({iv_shift.step} eV).'
                           ' Did you change the step of rpars.IV_SHIFT_RANGE '
                           'since the last R-factor calculation?') from exc
    return iv_shift


def _find_common_theory_experiment_range(theory, experiment,
                                         iv_shift, n_expand):
    """Return the range of theory energies in common with experiment."""
    # This is a little bit tricky. Here is what we want:
    #             |-----------------------|    theory
    #                     |-------------|      experiment
    #                     .             .
    #  |------------------vvvvv|        .      theory, shift max left
    #                |----vvvvvvvvvvvvvvv----| theory, shift max right
    # The 'v's mark regions of the shifted theory in common
    # with experiments, on the 'shifted' scales. (These are
    # the extremes. We want all the continuous shifts in
    # between too.) However, we want the union of all the
    # 'v' parts ON THE ORIGINAL ENERGY SCALE of the theory,
    # i.e.,
    #             |-----------------------|    theory
    #             |------------------vvvvv|    left, on same scale
    #             |----vvvvvvvvvvvvvvv----|    right, on same scale
    #             |----vvvvvvvvvvvvvvvvvvv|    Union of all shifts
    # Rather than doing this mess, we can equivalently EXPAND the
    # experimental range: on the left by the max right shift, on
    # the right by the max left shift, like so:
    #                     |-------------|            experiment
    #                  |RR|-------------|LLLLLLLLLL| expanded exp.
    #             |-----------------------|          theory
    #             |----vvvvvvvvvvvvvvvvvvv|          common range
    expanded_exp = experiment.copy()
    expanded_exp.start -= iv_shift.stop  # Expand when stop > 0
    expanded_exp.stop -= iv_shift.start  # Expand when start < 0

    # Expand by n_expand left & right, finally intersect
    theory_larger = theory.expanded_by(n_expand)
    try:
        return theory_larger.intersected(expanded_exp)
    except ValueError:  # No intersection
        raise RfactorError('Cannot run R-factor calculation: No common '
                           'energies between experiment and theory') from None


def sorted_energies_from_beams(beams):
    """Return a list of sorted energies from a list of Beam objects."""
    return sorted({e for beam in beams for e in beam.intens})


def writeWEXPEL(sl, rp, theobeams, filename="WEXPEL", for_error=False):
    """Write input file WEXPEL for R-factor calculation.

    Parameters
    ----------
    sl : Slab
        The Slab object containing atom information.
    rp : Rparams
        The run parameters.
    theobeams : list of Beam
        The theoretical beams, containing I(V) data.
    filename : str, optional
        Name of the file that will be written. The default is "WEXPEL".
    for_error : bool, optional
        Whether the R-factor calculation is used to determine error
        curves. The default is False.

    Returns
    -------
    None.
    """
    (_, theo_range,
     iv_shift, vincr) = prepare_rfactor_energy_ranges(rp, theobeams,
                                                      for_error,
                                                      n_expand=_N_EXPAND_THEO)

    # find correspondence experimental to theoretical beams:
    beamcorr = leedbase.getBeamCorrespondence(sl, rp)
    # integer & fractional beams
    iorf = []
    for (i, beam) in enumerate(rp.expbeams):
        if beam.hk[0] % 1.0 != 0.0 or beam.hk[1] % 1.0 != 0.0:
            iorf.append(2)
        else:
            iorf.append(1)
    iorf.extend([0]*(len(rp.ivbeams)-len(rp.expbeams)))

    f72 = ff.FortranRecordWriter('F7.2')
    if rp.TL_VERSION < Version('1.7.0'):
        beam_formatter = ff.FortranRecordWriter('25I3')
    else:
        beam_formatter = ff.FortranRecordWriter('25I4')
    i3 = ff.FortranRecordWriter('I3')

    # EMIN/EMAX are just two bounds to select which theory beams are
    #     read in. All energies outside the closed [EMIN, EMAX] range
    #     are not read in (and do not fill up any array). Notice that
    #     we take the maximum a bit larger, just to make sure that
    #     integer divisions in the FORTRAN code don't discard the
    #     upper limit.
    # EINCR is the interpolation step.
    # LIMFIL is number of consecutive input files. Hardcode to 1.
    # IPR is output formatting. Hardcode to 0.
    # **IMPORTANT**: VINCR should be an integer multiple of EINCR.
    # If the two are not multiples, the V0R shifts are wrong. Here,
    # for simplicity, we take them to be the exact same.
    output = f'''\
 &NL1
 EMIN={f72.write([theo_range.min]):>9},
 EMAX={f72.write([theo_range.max + 0.1 * vincr]):>9},
 EINCR={f72.write([vincr]):>8},
 LIMFIL=      1,
 IPR=         0,
 VI={f72.write([rp.V0_IMAG]):>11},
 V0RR=      0.0,
 V01={f72.write([iv_shift.start]):>10},
 V02={f72.write([iv_shift.stop]):>10},
 VINCR={f72.write([vincr]):>8},
 ISMOTH={i3.write([rp.R_FACTOR_SMOOTH]):>7},
 EOT=         0,
 PLOT=        1,
 GAP=         0,
 &END
'''
    output += beam_formatter.write([n+1 for n in beamcorr]) + '\n'
    if len(beamcorr) % 25 == 0:
        output += "\n"
    for i in range(0, 2):  # redundant since indices are already taken care of
        output += beam_formatter.write(
            [n+1 for n in range(len(rp.expbeams))]) + '\n'
        if len(rp.expbeams) % 25 == 0:
            output += '\n'
    output += beam_formatter.write(iorf) + '\n'
    if len(iorf) % 25 == 0:
        output += '\n'
    output += '&NL2\n'
    output += ' NSSK=    0,\n'
    if rp.R_FACTOR_TYPE == 1:
        output += ' WR=      0.,0.,1.,\n'  # Pendry
    elif rp.R_FACTOR_TYPE == 2:
        output += ' WR=      1.,0.,0.,\n'  # R2
    else:
        output += ' WR=      0.,1.,0.,\n'  # Zanazzi-Jona
    output += '''\
 &END
 &NL3
 NORM=           1,
 INTMAX=    999.99,
 PLSIZE=   1.0,1.0,
 XTICS=         50,
 &END
 '''
    auxexpbeams = writeAUXEXPBEAMS(rp.expbeams, header=rp.systemName,
                                   write=True, numbers=False)
    output += auxexpbeams + '\n'
    # information about gaps in the experimental spectra would go here
    try:
        with open(filename, 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error(f'Failed to write {filename}')
        raise
    logger.debug(f'Wrote to R-factor input file {filename} successfully')


def largest_nr_grid_points(rpars, theobeams, for_error,
                           n_expand=_N_EXPAND_THEO):
    """Return the largest possible number of grid points."""
    # The number of grid point is the quantity used to set up the
    # dimensions of the arrays that will contain the IV curves.
    # It should be at least as large as:
    # - the number of experiment energies, on the original grid
    # - the number of theory energies, on the original grid
    # - the number of experiment energies, on the interpolated grid
    # - the number of theory energies, on the interpolated grid
    # - in principle, there is also some relation to the number
    #   of energies in GAPS, if there's gaps in the data. Since
    #   we don't allow this, we do not care. Should we ever change
    #   our mind, we'd have to look more closely at the code.

    # Notice that, in principle, we would not need to keep extra room
    # for shifting the theory relative to experiments. The FORTRAN code
    # translates the 'shifted' indices to indices in the interpolated
    # version of the theory beams. However, it seems cleaner to use the
    # very same logic as in WEXPEL. This way we only need to maintain
    # one piece of code.
    (experiment, theory,
     _, interp_step) = prepare_rfactor_energy_ranges(rpars, theobeams,
                                                     for_error, n_expand)
    interp_exp = EnergyRange(experiment.start, experiment.stop, interp_step)
    section = 'refcalc' if not for_error else 'superpos'

    # Make sure to use the whole stored range of energies if available
    # (which corresponds also to what is in fd.out) and not just the
    # overlapping one!
    if rpars.theobeams[section]:
        interp_theo = EnergyRange(
            min([min(b.energies) for b in rpars.theobeams[section]]),
            max([max(b.energies) for b in rpars.theobeams[section]]),
            interp_step)
    else:
        interp_theo = EnergyRange(theory.start, theory.stop, interp_step)

    n_max = max((experiment.n_energies,
                 theory.n_energies,
                 interp_exp.n_energies,
                 interp_theo.n_energies))
    return round(np.ceil(n_max * 1.1))  # 10% headroom, just in case


def writeRfactPARAM(rp, theobeams, for_error=False, only_vary=None):
    """Generate the PARAM file for the R-factor calculation.

    Parameters
    ----------
    rp : Rparams
        The run parameters.
    theobeams : list of Beam
        The theoretical beams, containing I(V) data.
    for_error : bool, optional
        Whether this R-factor calculation is used to produce error
        curves. Default is False.
    only_vary : list or None, optional
        Which parameter should be varied. Items are SearchPar objects.
        This argument is used only if for_error. In that case, it is
        used solely to compute how many distinct parameters values
        are expected. If not given or None, all the parameters are
        used. Default is False.

    Raises
    ------
    RfactorError
        If this function is called before writeWEXPEL.
    """
    if not rp.IV_SHIFT_RANGE.has_step:
        raise RfactorError('Cannot writeRfactPARAM without interpolation '
                           'step. Did you forget to call writeWEXPEL first?')
    ngrid = largest_nr_grid_points(rp, theobeams, for_error)

    n_var = 1
    if for_error:
        if not only_vary:
            logger.warning('Rfactor PARAM for error: Parameters under '
                           'variation not passed.')
            only_vary = [sp for sp in rp.searchpars
                         if sp.atom in rp.search_atlist]
        n_var = max([sp.steps for sp in only_vary])
    output = f'''
C  MNBED  : number of beams in experimental spectra before averaging
C  MNBTD  : number of beams in theoretical spectra before averaging

      PARAMETER (MNBED = {len(rp.expbeams)}, MNBTD = {len(theobeams)})

C  MNET   : number of energies in theoretical beam at time of reading in
C  MNGP  : greater equal number of grid points in energy working grid (ie after
C           interpolation)
C  MNS    : number of geometries including those, that are skipped

      PARAMETER (MNET = {rp.THEO_ENERGIES.n_energies}, MNGP = {ngrid})
      PARAMETER (MNS = {n_var})

C  MNGAP  : number of gaps in the experimental spectra (NOTE: if there are no
C           gaps in the spectra set MNGAP to 1 to avoid zero-sized arrays)

      PARAMETER (MNGAP = 1)
'''
    # write PARAM
    try:
        with open('PARAM', 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error('Failed at writing PARAM file for R-factor calculation.')
        raise


def read_rfactor_columns(cols_dir=''):
    """
    Reads data from the theo.column and exp.column files in a given directory.

    Parameters
    ----------
    cols_dir : str, optional
        The directory to read from. The default is ''.

    Returns
    -------
    list [theo_beams, exp_beams]
        Both the theoretical and the experimental beams are formatted as lists
        of 2D numpy arrays, each of which has the form [[en1, intens1], ...]
    """

    fnames = ['theo.column', 'exp.column']
    xxyy = []
    for fname in fnames:
        try:
            f = open(os.path.join(cols_dir, fname), 'r')
        except FileNotFoundError:
            logger.error("read_rfactor_columns: File {} not found. Aborting."
                         .format(fname))
            return [], []
        except PermissionError:
            logger.error("read_rfactor_columns: Cannot open file {}. Aborting."
                         .format(fname))
            return [], []

        cols = np.array([[float(col) for col in line.split()] for line in f])
        f.close()
        xy = np.split(cols, np.shape(cols)[1]/2, axis=1)
        # xy is now a list of 2D arrays.
        # Each array has the form [[en1, intens1], ...]
        #
        # for each beam, get rid of the points that have (en, intens) = (0, 0)
        # so that they don't screw up the plots later
        xy = [coords[~np.all(coords < 1e-3, axis=1)] for coords in xy]
        if xy:
            xxyy.append(xy)
        else:
            logger.warning("File " + fname + " contains no usable data.")
    # xxyy now contains first the theoretical, then the experimental beams
    return xxyy


def writeRfactorPdf(beams, colsDir='', outName='Rfactor_plots.pdf',
                    analysisFile='', v0i=0., formatting=None):
    '''
    Creates a single PDF file containing the plots of R-factors, using plot_iv.
    If analysisFile is defined, a second 'analysis' PDF will be generated.

    Parameters
    ----------
    beams : list of (name, R) tuples
           name: str
                 formatted fractional index of the beam
           R: float
              R-factor value
    colsDir : kwarg, str
        path to folder containing the files theo.column and exp.column
        generated by the R-factor routine of TensErLEED.
        default: current path
    outName : kwarg, str
        name of the file (with or without extension) to which the plots
        will be saved.
        default: 'Rfactor_plots.pdf'
    analysisFile : kwarg, string
        if not empty, a more extensive R-factor analysis pdf with
        calculated Y-functions and absolute errors will be written to the
        given file name.
    v0i : kwarg, float
        imaginary part of the inner potential for calculating Y-functions.
        Should always be passed if analysisFile is passed.
    formatting : kwarg, dict
        dict containing formatting instructions:
        axes : str
            Which axes to draw.
            all:  draw all axes (left, right, top, bottom) for all panels
            lb:   draw only left and bottom axes
            b:    draw only bottom axes
            none: draw no axes except at bottom of last panel in each column
        colors : tuple (str, str)
            Define alternative colors for experimental and theoretical beams.
            default: None
        legend : str
            Which legends to print.
            all:   print legend for each panel
            first: print legend only for the first panel on each page
            tr:    print legend only for the top-right panel on each page
            none:  do not print any legend.
        perpage : int or (int, int)
            Define how many figures to plot on each page. Either tuple
            (columns, rows), or single integer (preferred). For single int,
            layout will be adapted automatically. Numbers that are not nicely
            divisible may be rounded up, resulting in some whitespace.
            default: 2 (one column, two rows).

    Returns
    -------
    None

    '''

    global _CAN_PLOT
    if not _CAN_PLOT:
        logger.debug("Necessary modules for plotting not found. Skipping "
                     "R-factor plotting.")
        return

    xyTheo, xyExp = read_rfactor_columns(cols_dir=colsDir)
    labels, rfacs = zip(*beams)
    rfac_str = ["R = {:.4f}".format(r) for r in rfacs]
    plot_iv([xyTheo, xyExp], outName, legends=['Theoretical', 'Experimental'],
            labels=labels, annotations=rfac_str, formatting=formatting)

    if not analysisFile:
        return

    figs, figsize, namePos, oritick, plotcolors, rPos, xlims, ylims = prepare_analysis_plot(formatting, xyExp, xyTheo)

    try:
        pdf = PdfPages(analysisFile)
    except PermissionError:
        logger.error("writeRfactorPdf: Cannot open file {}. Aborting."
                     .format(analysisFile))
        return
    # the following will spam the logger with debug messages; disable.
    loglevel = logger.level
    logger.setLevel(logging.INFO)
    try:
        for i, (name, rfact, theo, exp) in enumerate(zip(*zip(*beams),
                                                         xyTheo, xyExp)):

            if len(exp) == 0:
                continue
            ytheo = leedbase.getYfunc(theo, v0i)
            yexp = leedbase.getYfunc(exp, v0i)

            plot_analysis(exp, figs, figsize, name, namePos, oritick,
                          plotcolors, rPos, rfact, theo, xlims, yexp,
                          ylims, ytheo, v0i)

        for fig in figs:
            pdf.savefig(fig)
            plt.close(fig)
    except Exception:
        logger.error("writeRfactorPdf: Error while writing analysis pdf: ",
                     exc_info=True)
    finally:
        pdf.close()
        logger.setLevel(loglevel)
    return


def prepare_analysis_plot(formatting, xyExp, xyTheo):
    # write R-factor analysis
    # find min and max values of x and y for plotting all curves
    # on the same horizontal scale and leaving a little y space for the legend
    xmin = min(min(xy[:, 0]) for xy in [*xyTheo, *xyExp] if len(xy) != 0)
    xmax = max(max(xy[:, 0]) for xy in [*xyTheo, *xyExp] if len(xy) != 0)
    ymin = min(min(xy[:, 1]) for xy in [*xyTheo, *xyExp] if len(xy) != 0)
    ymax = max(max(xy[:, 1]) for xy in [*xyTheo, *xyExp] if len(xy) != 0)
    dy = ymax - ymin
    # set ticks spacing to 50 eV and round the x limits to a multiple of it
    tick = 50
    oritick = plticker.MultipleLocator(base=tick)
    xlims = (np.floor(xmin / tick) * tick,
             np.ceil(xmax / tick) * tick)
    dx = xlims[1] - xlims[0]
    # set formatting parameters
    plotcolors = []
    if formatting is not None:
        if 'colors' in formatting:
            plotcolors = formatting['colors']
    figsize = (5.8, 8.3)
    figs = []
    ylims = (ymin - 0.02 * dy, ymax + 0.22 * dy)
    namePos = (xlims[0] + 0.45 * dx, ylims[1] - 0.1 * dy)
    rPos = (namePos[0], namePos[1] - 0.085 * dy)
    return figs, figsize, namePos, oritick, plotcolors, rPos, xlims, ylims


def plot_analysis(exp, figs, figsize, name, namePos, oritick, plotcolors, rPos, rfact, theo, xlims, yexp, ylims, ytheo, v0i):
    fig, axs = plt.subplots(3, figsize=figsize,
                            squeeze=True)
    fig.subplots_adjust(left=0.06, right=0.94,
                        bottom=0.07, top=0.98,
                        wspace=0, hspace=0.08)
    figs.append(fig)
    [ax.set_xlim(*xlims) for ax in axs]
    axs[0].set_ylim(*ylims)
    if v0i is not None: # if is None, don't scale
        # Y function plot limits:
        # Y osciallates between +/- 1/(2*abs(v0i))
        y_range_uni_direc = 1/(2*abs(v0i)) * 1.1 # 10% spacing on either side
        axs[1].set_ylim(-1*y_range_uni_direc, y_range_uni_direc)
    [ax.get_yaxis().set_ticks([]) for ax in axs]
    [ax.tick_params(bottom=True,
                    top=True,
                    axis='x', direction='in') for ax in axs]
    [ax.xaxis.set_major_locator(oritick) for ax in axs]
    axs[0].set_ylabel("Intensity (arb. units)")
    axs[1].set_ylabel("Y")
    axs[2].set_ylabel(r'$\sum\Delta Y^2 / \left(Y_1^2 + Y_2^2\right)$')
    axs[2].set_xlabel("Energy (eV)")
    y_theo_sq = np.copy(ytheo)
    y_theo_sq[:, 1] = y_theo_sq[:, 1] ** 2
    y_exp_sq = np.copy(yexp)
    y_exp_sq[:, 1] = y_exp_sq[:, 1] ** 2
    dy = np.array([(ytheo[j, 0], yexp[j, 1] - ytheo[j, 1])
                   for j in range(0, min(len(ytheo), len(yexp)))])
    dysq = np.copy(dy)
    dysq[:, 1] = dysq[:, 1] ** 2

    eps = 1e-10 # needed to avoid division by 0. Only important for plotting.
    norm_y_squares = np.array(
        [[dysq[j, 0], (dysq[j, 1]
                       / (y_theo_sq[j, 1] + y_exp_sq[j, 1] + eps))]
         for j in range(len(dysq))])
    sum_norm_y_squares = np.array([norm_y_squares[0]])
    for j in range(1, len(norm_y_squares)):
        sum_norm_y_squares = (
            np.append(sum_norm_y_squares,
                      [[norm_y_squares[j, 0],
                        (sum_norm_y_squares[j - 1, 1]
                         + norm_y_squares[j, 1])]],
                      axis=0))
    axs[1].plot(xlims, [0., 0.], color='grey', alpha=0.2)
    if not plotcolors:
        if not all([matplotlib.colors.is_color_like(s)
                    for s in plotcolors]):
            plotcolors = []
            logger.warning("writeRfactorPdf: Specified colors not "
                           "recognized, reverting to default colors")
    if not plotcolors:
        axs[0].plot(theo[:, 0], theo[:, 1], label='Theoretical')
        axs[0].plot(exp[:, 0], exp[:, 1], label='Experimental')
        axs[1].plot(ytheo[:, 0], ytheo[:, 1], label='Theoretical')
        axs[1].plot(yexp[:, 0], yexp[:, 1], label="Experimental")
    else:
        axs[0].plot(theo[:, 0], theo[:, 1], label='Theoretical',
                    color=plotcolors[0])
        axs[0].plot(exp[:, 0], exp[:, 1], label='Experimental',
                    color=plotcolors[1])
        axs[1].plot(ytheo[:, 0], ytheo[:, 1], label='Theoretical',
                    color=plotcolors[0], linewidth=0.75)
        axs[1].plot(yexp[:, 0], yexp[:, 1], label="Experimental",
                    color=plotcolors[1], linewidth=0.75)
    axs[1].plot(dy[:, 0], dy[:, 1], label="\u0394Y", color="black",
                linewidth=0.5)
    axs[1].fill_between(dy[:, 0], dy[:, 1], 0., facecolor='grey',
                        alpha=0.5)
    axs[2].plot(sum_norm_y_squares[:, 0], sum_norm_y_squares[:, 1],
                color="black", drawstyle="steps-mid")
    axs[0].annotate(name, namePos, fontsize=10)
    axs[0].annotate("R = {:.4f}".format(rfact), rPos, fontsize=10)
    axs[0].legend()
    axs[1].legend()


def writeRfactorPdf_new(n_beams, labels, rfactor_beams,
                        energies, id_start,
                        n_E_beams,
                        int_1, int_2, y_1, y_2 ,
                        outName='Rfactor_plots.pdf',
                        analysisFile='', v0i = 0., formatting=None):

    # after applying the V0r shift outside, the id_start and n_E_beams should be same for experiment and theory
    global _CAN_PLOT
    if not _CAN_PLOT:
        logger.debug("Necessary modules for plotting not found. Skipping "
                     "R-factor plotting.")
        return


    # get data
    exp_xy = []
    theo_xy = []
    for i in range(n_beams):
        xy = np.empty([n_E_beams[i], 2])
        xy[:, 0] = energies[id_start[i] -1: id_start[i] + n_E_beams[i] -1]
        xy[:, 1] = int_1[id_start[i] -1: id_start[i] + n_E_beams[i] -1, i]
        # normalize to max of beam:
        xy[:, 1] /= np.nanmax(xy[:, 1])
        exp_xy.append(xy)


        xy = np.empty([n_E_beams[i], 2]) # want this at same range as exp only!
        xy[:, 0] = energies[id_start[i] -1: id_start[i] + n_E_beams[i] -1]
        xy[:, 1] = int_2[id_start[i] -1: id_start[i] + n_E_beams[i] -1, i]
        # normalize to max of beam:
        xy[:, 1] /= np.nanmax(xy[:, 1])
        theo_xy.append(xy)

    data = [theo_xy, exp_xy]

    rfac_str = ["R = {:.4f}".format(r) for r in rfactor_beams]
    plot_iv(data, outName, legends=['Theoretical', 'Experimental'],
            labels=labels, annotations=rfac_str, formatting=formatting)

    if not analysisFile:
        return

    figs, figsize, namePos, oritick, plotcolors, rPos, xlims, ylims = \
        prepare_analysis_plot(formatting, exp_xy, theo_xy)


    try:
        pdf = PdfPages(analysisFile)
    except PermissionError:
        logger.error("writeRfactorPdf: Cannot open file {}. Aborting."
                     .format(analysisFile))
        return
    # the following will spam the logger with debug messages; disable.
    loglevel = logger.level
    logger.setLevel(logging.INFO)

    # proper minus character
    labels = [label.replace("-", "−") for label in labels]

    try:
        for i in range(n_beams):
            exp = exp_xy[i]
            theo = theo_xy[i]
            beam_energies = energies[id_start[i] -1: id_start[i] + n_E_beams[i] -1]
            y_exp = np.empty([n_E_beams[i], 2])
            y_theo = np.empty([n_E_beams[i], 2])
            y_exp[:, 0] = beam_energies
            y_exp[:, 1] = y_1[id_start[i] -1: id_start[i] + n_E_beams[i] -1, i]
            y_theo[:, 0] = beam_energies
            y_theo[:, 1] = y_2[id_start[i] -1: id_start[i] + n_E_beams[i] -1, i]

            plot_analysis(exp, figs, figsize, labels[i], namePos, oritick, plotcolors, rPos, rfactor_beams[i],
                          theo, xlims, y_exp, ylims, y_theo, v0i)

        for fig in figs:
            pdf.savefig(fig)
            plt.close(fig)
    except Exception:
        logger.error("writeRfactorPdf: Error while writing analysis pdf: ",
                     exc_info=True)
    finally:
         pdf.close()
         logger.setLevel(loglevel)
    return


def beamlist_to_array(beams):
    # turn list of Beam objects into an array of intensities

    n_beams = len(beams)
    energies = sorted_energies_from_beams(beams)
    in_grid = np.array(energies)
    n_E = in_grid.shape[0]

    # fill with NaNs as default value
    beam_arr = np.full([n_E, n_beams], fill_value=np.NaN)

    id_start = np.zeros(n_beams, dtype=np.int32)
    n_E_beams = np.zeros(n_beams, dtype=np.int32)

    for i, b in enumerate(beams):
        # write beams into colums of beam_arr
        id_start[i] = np.where(np.isclose(energies, min(b.intens.keys())))[0][0]
        n_E_beams[i] = len(b.intens)
        beam_arr[id_start[i]: id_start[i] + n_E_beams[i], i] = list(b.intens.values())


    return in_grid, id_start, n_E_beams, beam_arr




"""=================
    ViPErLEED
=================
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-01'
__license__ = 'GPLv3+'

from viperleed.cli import ViPErLEEDMain

ViPErLEEDMain.run_as_script()
